"""
    Copyright 2015-2021 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from .. import eggwrapper as _eggwrapper
from .. import stats, _interface

class ParamDict(object):
    """
    :class:`dict`-like class managing parameters. Do most of what a
    dictionary does except for adding and removing keys. The order of
    parameters is fixed and consistent.

    In addition to the methods documented below, :class:`.ParamDict`
    instances support the following operations (where ``params`` in one
    instance):

    +------------------------+--------------------------------------------------------+
    |Expression              | Action                                                 |
    +========================+========================================================+
    |``len(params)``         | Number of parameters                                   |
    +------------------------+--------------------------------------------------------+
    |``params[key]``         | Get the value of parameter ``key``                     |
    +------------------------+--------------------------------------------------------+
    |``params[key] = value`` | Assign ``value`` to parameter ``key`` (see note below) |
    +------------------------+--------------------------------------------------------+
    |``for key in params``   | Same as ``for key in params.keys()``                   |
    +------------------------+--------------------------------------------------------+
    |``reversed(params)``    | Reversed iterator                                      |
    +------------------------+--------------------------------------------------------+
    |``key in params``       | Check if ``key`` in a parameter name                   |
    +------------------------+--------------------------------------------------------+
    |``str(params)``         | Representation of the instance as a :class:`dict`      |
    +------------------------+--------------------------------------------------------+

    Note that the ``params[key] = value`` expression is straightforward
    only for parameters that have a single value. For parameters
    represented by a :class:`.ParamList` instance, this expression is
    only supported if the right-hand operand is a sequence of matching
    length (in order to set all values at once). For parameters
    represented by a  :class:`.ParamMatrix` instance, this expression is
    only supported if the right-hand operand is another
    :class:`.ParamMatrix` instance of matching dimension. For
    ``events``, this expression is not supported at all. In all cases
    where ``params[key]`` returns :class:`.ParamList`,
    :class:`.ParamMatrix` or a :class:`.EventList` instance, the
    returned value can be modified using its own methods.
    """
    _keys = ['num_pop', 'num_sites', 'recomb', 'theta', 'num_mut',
             'mut_model', 'TPM_proba', 'TPM_param', 'num_alleles',
             'rand_start', 'num_chrom', 'num_indiv', 'N', 'G', 's',
             'site_pos', 'site_weight', 'migr_matrix', 'trans_matrix',
             'events', 'max_iter']

    # sub-lists with only parameters that can be changed between
    # simulations
    _keys_scalar = ['num_sites', 'recomb', 'theta', 'num_mut',
             'mut_model', 'TPM_proba', 'TPM_param', 'num_alleles',
             'rand_start']
    _keys_list = ['num_chrom', 'num_indiv', 'N', 'G', 's', 'site_pos',
                  'site_weight']
    _keys_matrix = ['migr_matrix', 'trans_matrix']

    def __init__(self, npop):
        if not isinstance(npop, int): raise TypeError('invalid type for `npop`')
        if npop < 1: raise ValueError('invalid value for `npop`')

        self._params = _eggwrapper.Params(npop, 0.0)
        self._migr = self._params.M()
        self._params.set_L(0)
        self._npop = npop

        self._num_chrom = ParamList(getter=self._params.get_n1,
                                    setter=self._params.set_n1,
                                    num_values=npop,
                                    check_item=lambda x: x>=0)

        self._num_indiv = ParamList(getter=self._params.get_n2,
                                    setter=self._params.set_n2,
                                    num_values=npop,
                                    check_item=lambda x: x>=0)

        self._N = ParamList(getter=self._params.get_N,
                            setter=self._params.set_N,
                            num_values=npop,
                            check_item=lambda x: x>0.0)

        self._G = ParamList(getter=self._params.get_G,
                            setter=self._params.set_G,
                            num_values=npop,
                            check_item=lambda x: True)

        self._s = ParamList(getter=self._params.get_s,
                            setter=self._params.set_s,
                            num_values=npop,
                            check_item=lambda x: x>=0.0 and x<=1.0)

        self._site_pos = ParamList(getter=self._params.get_sitePos,
                                   setter=self._params.set_sitePos,
                                   num_values=0,
                                   check_item=lambda x: x>=0.0 and x<=1.0)

        self._site_weight = ParamList(getter=self._params.get_siteW,
                                      setter=self._params.set_siteW,
                                      num_values=0,
                                      check_item=lambda x: x>0.0)

        self._migr_matrix = ParamMatrix(getter=self._migr.get_pair,
                                        setter=self._migr.set_pair,
                                        num_values=npop,
                                        check_item=lambda x:x>=0.0,
                                        check_active=lambda: True,
                                        set_active=None)

        self._trans_matrix = ParamMatrix(getter=self._params.get_transW_pair,
                                         setter=self._params.set_transW_pair,
                                         num_values=2,
                                         check_item=lambda x:x>0.0,
                                         check_active=lambda: self._params.get_transW_matrix(),
                                         set_active=self._params.set_transW_matrix)

        self._events = EventList(npop=npop,
                                 add=self._params.addChange,
                                 clear=self._params.clearChanges)

    def disable_trans_matrix(self):
        """
        Disable transition weight matrix.
        Disable the matrix of weights of transition between alleles (it
        is disabled by default, and automatically activated i any value
        is set). This sets all weights to 1.0.
        """
        self._params.set_transW_matrix(False)

    def get_values(self, other):
        """
        Import parameter values.
        Update the dictionary with values from another
        :class:`.ParamDict` instance (with the same numbers of
        populations and sites).
        """
        if not isinstance(other, ParamDict): raise TypeError('`other` should be a `ParamDict` instance')
        self['num_sites'] = other['num_sites']
        self['num_alleles'] = other['num_alleles']
        for k in self._keys:
            if k not in ['num_pop', 'num_sites', 'num_alleles', 'events']: self[k] = other[k]
        self._events.replace(other._events)

    def summary(self):
        """
        Parameter summary.
        Return a string displaying all current values of parameters at
        the level of the C++ simulator. Use for debugging only, as the
        format is not guaranteed to be stable.
        """
        return self._params.summary()

    def _set(self, key, values): # alternative setter
        if key in self._keys_scalar: self[key] = values
        elif key in self._keys_list: self[key][values[0]] = values[1]
        elif key in self._keys_matrix: self[key][values[0], values[1]] = values[2]
        else: raise KeyError('this parameter cannot be modified between simulations')

    def set_migr(self, value):
        """
        Set migration rates.
        Set all pairwise migration rates to ``value/(num_pop-1)``.
        """
        if value < 0.0: raise ValueError('invalid value for `migr`')
        self._migr.set_all(value)

    def mk_structure(self, skip_indiv=False, outgroup_label=None):
        """
        Create structure object.
        Export a :class:`.Structure` instance containing the
        structure information corresponding to currently loaded
        simulation parameters.

        :param skip_indiv: this argument determines whether the individuals
            level should be skipped (if ``True``, alleles from each given
            individual are treated as belonging to separate haploid individuals.
        
        :param outgroup_label: this argument indicates the label of the outgroup
            population. It must be passed as a string. Default value is None.

            .. warning::
                The individual level cannot be processed if the ploidy is
                not constant (mixing sampled chromosomes and sampled
                individuals).
        
        :return: A new :class:`.Structure` instance.
        """

        # process populations
        struct_ing = {None: {}}
        struct_out = {}
        cnt_sam = 0
        cnt_idv = 0
        
        for k in range(self._npop):

            # get list of sample indexes (as a list of tuples/per individual)
            sam_start = cnt_sam
            idv = [(sam_start+i*2, sam_start+i*2+1) for i in range(self._params.get_n2(k))]
            sam_start += 2*len(idv)
            idv.extend([(sam_start+i,) for i in range(self._params.get_n1(k))])
            cnt_sam += 2*self._params.get_n2(k) + self._params.get_n1(k)

            # process population
            if skip_indiv:
                idv = [(j,) for i in idv for j in i]
            idv = dict(zip(map(str, range(cnt_idv, cnt_idv+len(idv))), idv))
            
            if outgroup_label != None and k == int(outgroup_label):
                struct_out = idv
            else:
                struct_ing[None][str(k)] = idv
            
            cnt_idv += len(idv)

        # process delayed samples
        if self._params.nDSChanges() > 0:
            for event in self._events:
                if event['cat'] == 'sample':
                    k = event['label']
                    idv = [(cnt_sam+i*2, cnt_sam+i*2+1) for i in range(event['num_indiv'])]
                    idv.extend([(cnt_sam+i,) for i in range(event['num_chrom'])])
                    cnt_sam += 2 * event['num_indiv'] + event['num_chrom']
                    
                    if skip_indiv: idv = [(j,) for i in idv for j in i]
                    idv = dict(zip(map(str, range(cnt_idv, cnt_idv+len(idv))), idv))
                    
                    if outgroup_label != None and k == int(outgroup_label):
                        struct_out = idv
                    else:
                        if str(k) not in struct_ing[None]: struct_ing[None][str(k)] = {}
                        struct_ing[None][str(k)].update(idv)
                    cnt_idv += len(idv)
        
        if len(struct_ing[None]) == 0: raise ValueError('cannot generate structure')
        return _interface.struct_from_dict(struct_ing, struct_out)

    ####################################################################
    # below: dict-like methods
    ####################################################################

    def has_key(self, key):
        """
        Tell if a parameter exists.
        Return a boolean indicating if the passed name is one of the
        parameters.
        """
        return key in self._keys

    def get(self, key, default=None):
        """
        Get a parameter value.
        Return the value for *key* if *key* is one of the parameters,
        else return the value passed as *default* (by default,
        ``None``). This method therefore never raises a
        :exc:`KeyError`.
        """
        if key in self._keys: return self[key]
        return default

    def __iter__(self):
        for key in self._keys: yield key

    def keys(self):
        """
        Iterator over the parameter names.
        """
        for key in self._keys: yield key

    def values(self):
        """
        Iterator over the parameter values.
        """
        for key in self._keys: yield self[key]

    def items(self):
        """
        Iterator over content of the instance. Each iteration round
        yield a ``(key, value)`` parameter name and value pair.
        """
        for key in self._keys: yield key, self[key]

    def copy(self):
        """
        Shallow copy.
        Return a shallow copy of this instance. All modifications of the
        values of either copy will modify the other one (even
        modification of integer, float and string parameters).
        """
        ret = ParamDict(self._npop)
        ret._params = self._params
        return ret

    def update(self, other=None, **kwargs):
        """
        Import parameter values.
        Update the dictionary with the ``(key, value)`` parameter name
        and value pairs from the object passed as *other*, overwriting
        existing keys and raising a :exc:`KeyError`
        exception if an unknown parameter name is met.

        Accepts :class:`dict` instances, or else any iterable of
        ``(key, value)`` pairs. If keyword arguments are specified, the
        instance is then updated with those, as in
        ``param_dict.update(theta=5.0, recomb=2.5)``.
        """
        if isinstance(other, ParamDict):
            self.get_values(other)
        elif other is not None:
            if 'num_sites' in other: self['num_sites'] = other['num_sites']
            if 'num_alleles' in other: self['num_alleles'] = other['num_alleles']
            try:
                iter_ = other.items()
            except AttributeError: iter_ = iter(other)
            for k, v in iter_: self[k] = v
        if 'num_sites' in kwargs:
            self['num_sites'] = kwargs['num_sites']
            del kwargs['num_sites']
        if 'num_alleles' in kwargs:
            self['num_alleles'] = kwargs['num_alleles']
            del kwargs['num_alleles']
        for k, v in kwargs.items(): self[k] = v

    def __len__(self):
        return len(self._keys)

    def __reversed__(self):
        for key in reversed(self._keys):
            yield key

    def __contains__(self, key):
        return key in self._keys

    def __str__(self):
        return str(dict(self.items()))

    def __getitem__(self, key):
        if key == 'num_pop': return self._npop
        elif key == 'num_sites': return self._params.get_L()
        elif key == 'recomb': return self._params.get_R()
        elif key == 'theta': return self._params.get_theta()
        elif key == 'num_mut': return self._params.get_fixed()
        elif key == 'mut_model':
            model = self._params.get_mutmodel()
            if model == _eggwrapper.Params.KAM: return 'KAM'
            elif model == _eggwrapper.Params.IAM: return 'IAM'
            elif model == _eggwrapper.Params.SMM: return 'SMM'
            elif model == _eggwrapper.Params.TPM: return 'TPM'
        elif key == 'TPM_proba': return self._params.get_TPMproba()
        elif key == 'TPM_param': return self._params.get_TPMparam()
        elif key == 'num_alleles': return self._params.get_K()
        elif key == 'rand_start': return self._params.get_random_start_allele()
        elif key == 'num_chrom': return self._num_chrom
        elif key == 'num_indiv': return self._num_indiv
        elif key == 'N': return self._N
        elif key == 'G': return self._G
        elif key == 's': return self._s
        elif key == 'site_pos': return self._site_pos
        elif key == 'site_weight': return self._site_weight
        elif key == 'migr_matrix': return self._migr_matrix
        elif key == 'trans_matrix': return self._trans_matrix
        elif key == 'events': return self._events
        elif key == 'max_iter': return self._params.get_max_iter()
        else: raise KeyError('invalid parameter name: {0}'.format(key))

    def __setitem__(self, key, value):
        if key == 'num_pop': raise ValueError('the number of populations is read-only (it must be set at construction time)')
        elif key == 'num_sites':
            if value < 0: raise ValueError('invalid value for `{0}`'.format(key))
            self._params.set_L(value)
            self._site_pos._num_values = value
            self._site_weight._num_values = value
            if value > 0: self._params.autoSitePos()
        elif key == 'recomb':
            if value < 0.0: raise ValueError('invalid value for `{0}`'.format(key))
            self._params.set_R(value)
        elif key == 'theta':
            if value < 0: raise ValueError('invalid value for `{0}`'.format(key))
            if value > 0.0 and self._params.get_fixed() > 0.0: raise ValueError('it is not allowed to set both theta and the number of mutations to non-zero')
            self._params.set_theta(value)
        elif key == 'num_mut':
            if value < 0: raise ValueError('invalid value for `{0}`'.format(key))
            if value > 0.0 and self._params.get_theta() > 0.0: raise ValueError('it is not allowed to set both theta and the number of mutations to non-zero')
            self._params.set_fixed(value)
        elif key == 'mut_model':
            if value == 'KAM': self._params.set_mutmodel(_eggwrapper.Params.KAM)
            elif value == 'IAM': self._params.set_mutmodel(_eggwrapper.Params.IAM)
            elif value == 'SMM': self._params.set_mutmodel(_eggwrapper.Params.SMM)
            elif value == 'TPM': self._params.set_mutmodel(_eggwrapper.Params.TPM)
            else: raise ValueError('invalid value for `{0}`'.format(key))
        elif key == 'TPM_proba':
            if value < 0 or value > 1: raise ValueError('invalid value for `{0}`'.format(key))
            self._params.set_TPMproba(value)
        elif key == 'TPM_param':
            if value < 0 or value > 1: raise ValueError('invalid value for `{0}`'.format(key))
            self._params.set_TPMparam(value)
        elif key == 'num_alleles':
            if value < 2: raise ValueError('invalid value for `{0}`'.format(key))
            self._params.set_K(value)
            self._trans_matrix._num_values = value
        elif key == 'rand_start':
            self._params.set_random_start_allele(value)
        elif key == 'max_iter':
            if value < 0: raise ValueError('maximum number of iterations cannot be negative')
            self._params.set_max_iter(value)
        elif key == 'num_chrom': self._num_chrom[:] = value
        elif key == 'num_indiv': self._num_indiv[:] = value
        elif key == 'N': self._N[:] = value
        elif key == 'G': self._G[:] = value
        elif key == 's': self._s[:] = value
        elif key == 'site_pos': self._site_pos[:] = value
        elif key == 'site_weight': self._site_weight[:] = value
        elif key == 'migr_matrix': self._migr_matrix.get_values(value)
        elif key == 'trans_matrix': self._trans_matrix.get_values(value)
        elif key == 'events': raise ValueError('cannot replace the list of events as a whole')
        else: raise KeyError('invalid parameter name: `{0}`'.format(key))

    def add_event(self, cat, T, **params):
        """
        Add an event to the list. See the documentation for the class
        :class:`.Simulator` for more details on what is expected as
        arguments.

        :param cat: category of the event.
        :param T: date of the event.
        :param params: all needed parameters for this category of
            events.
        """
        self._events.add(cat, T, **params)

class ParamList(object):
    """
    :class:`list`-like class managing values for a parameter. Do
    most of what a list does except for adding and removing values. No
    methods :meth:`!reverse` and :meth:`!sort` are provided either
    as they make little sense. Expressions involving the ``+``
    and ``*`` arithmetic operators are supported, but return genuine
    :class:`list` instances that are disconnected from the original
    parameter holder.

    In addition to the methods documented below, :class:`.ParamList`
    instances support the following operations (where ``items`` is a
    :class:`.ParamList` instance):

    +--------------------------+----------------------------------------------------------------+
    |Expression                | Action                                                         |
    +==========================+================================================================+
    |``len(items)``            | Number of items                                                |
    +--------------------------+----------------------------------------------------------------+
    |``items[i]``              | Get *i*\ th item                                               |
    +--------------------------+----------------------------------------------------------------+
    |``items[i:j]``            | Slice from *i* to *j*                                          |
    +--------------------------+----------------------------------------------------------------+
    |``items[i:j:k]``          | Slice from *i* to *j* with step *k*                            |
    +--------------------------+----------------------------------------------------------------+
    |``items[i] = value``      | Assign ``value`` to parameter ``key``                          |
    +--------------------------+----------------------------------------------------------------+
    |``items[i:j] = values``   | Replace a slice of values (with a sequence of the same length) |
    +--------------------------+----------------------------------------------------------------+
    |``items[i:j:k] = values`` | Replace a slice of values (with a sequence of the same length) |
    +--------------------------+----------------------------------------------------------------+
    |``for item in items``     | Iterates over items                                            |
    +--------------------------+----------------------------------------------------------------+
    |``reversed(items)``       | Reversed iterator                                              |
    +--------------------------+----------------------------------------------------------------+
    |``item in items``         | Check if ``item`` is present among items                       |
    +--------------------------+----------------------------------------------------------------+
    |``items + other``         | Same as ``list(items) + other`` (returns a :class:`list`)      |
    +--------------------------+----------------------------------------------------------------+
    |``items * n``             | Same as ``list(items) * n`` (returns a :class:`list`)          |
    +--------------------------+----------------------------------------------------------------+
    |``str(items)``            | Representation of the instance as a list                       |
    +--------------------------+----------------------------------------------------------------+

    The indexing operator (``[]``) supports negative indexes (to count
    from the end) and slices operators, just as for the built-in type
    :class:`list`. However, all operators (such as ``del``) or methods
    (such as :meth:`!append`, :meth:`!extend`,
    :meth:`!remove`) that would change the length of the list are
    not available.
    """

    def __init__(self, getter, setter, num_values, check_item):
        self._getter = getter
        self._setter = setter
        self._num_values = num_values
        self._check_item = check_item

    def __getitem__(self, index):
        if isinstance(index, slice):
            return tuple(map(self.__getitem__, range(*index.indices(self._num_values))))
        else:
            if index < 0: index = self._num_values + index
            if index >= self._num_values: raise ValueError('list index out of range')
            return self._getter(index)

    def __setitem__(self, index, value):
        if isinstance(index, slice):
            rng = list(range(*index.indices(self._num_values)))
            if len(rng) != len(value): raise ValueError('number of values is required to match number of indexes for a slice assignment')
            for i, v in zip(rng, value):
                if self._check_item(v) == False: raise ValueError('value out of range')
                self._setter(i, v)
        else:
            if index < 0: index = self._num_values + index
            if index >= self._num_values: raise ValueError('list index out of range')
            if self._check_item(value) == False: raise ValueError('value out of range')
            self._setter(index, value)

    def __iter__(self):
        for i in range(self._num_values):
            yield self._getter(i)

    def count(self, value):
        """
        Number of items that are equal to *value*.
        """
        cnt = 0
        for i in range(self._num_values):
            if self._getter(i) == value: cnt += 1
        return cnt

    def index(self, value, imin=0, imax=None):
        """
        Index of the first value matching *value*. The
        returned value is ``>=imin`` and ``<imax`` if they are provided.
        Raise a :exc:`ValueError` if the value is not found.
        """
        if imax is None: imax = self._num_values
        for i in range(imin, imax):
            if self._getter(i) == value: return i
        else:
            raise ValueError('{0} is not in list'.format(value))

    def __add__(self, other):
        return list(self) + other

    def __radd__(self, other):
        return other + list(self)

    def __mul__(self, num):
        return list(self) * num

    def __rmul__(self, num):
        return num * list(self)

    def __contains__(self, value):
        for i in range(self._num_values):
            if self._getter(i) == value: return True
        else:
            return False

    def __len__(self):
        return self._num_values

    def __str__(self):
        return str(list(self))

    def __repr__(self):
        if self._num_values <= 10: return str(list(self))
        else: return '<list of {0} values>'.format(self._num_values)

    def __reversed__(self):
        for i in reversed(range(self._num_values)):
            yield self._getter(i)

class ParamMatrix(object):
    """
    :class:`list`-like class managing values for a parameter as a
    matrix. Similar to :class:`.ParamList` except that it holds a
    matrix. It therefore implements less methods (no support for ``+``
    and ``*`` operators, no slices) and use double indexing system as in
    ``matrix[i, j] = x``, allowing both setting and getting values. The
    iterator, such as in ``for i in matrix``, flattens the matrix,
    returning all values from the first row, then those from the second
    tow, and so on (replacing the diagonal values by ``None``).
    ``len(matrix)`` returns the dimension (number of rows, which is the
    same as the number of columns).

    In addition to the methods documented below, :class:`.ParamMatrix`
    instances support the following operations (where ``matrix`` is a
    :class:`.ParamMatrix` instance):

    +--------------------------+----------------------------------------------------------------------+
    |Expression                | Action                                                               |
    +==========================+======================================================================+
    |``len(matrix)``           | Size of the matrix (as number of rows/colums)                        |
    +--------------------------+----------------------------------------------------------------------+
    |``matrix[i,j]``           | Get *j*\ th item of the *i*\ th row (``None`` for diagonal)          |
    +--------------------------+----------------------------------------------------------------------+
    |``matrix[i,j] = value``   | Assign ``value`` to *(i,j)*\ th item (no diagonal)                   |
    +--------------------------+----------------------------------------------------------------------+
    |``for item in matrix``    | Iterates over items (row-first *flat* iteration)                     |
    +--------------------------+----------------------------------------------------------------------+
    |``reversed(matrix)``      | Reversed iterator                                                    |
    +--------------------------+----------------------------------------------------------------------+
    |``item in matrix``        | Check if ``item`` is present among items                             |
    +--------------------------+----------------------------------------------------------------------+
    |``str(matrix)``           | Representation of the instance as a nested list (including diagonal) |
    +--------------------------+----------------------------------------------------------------------+

    It is not possible to slice-set ranges of values in the matrix, but
    one can set all values at once from a compatible source with
    :meth:`~.ParamMatrix.get_values`.
    """

    def __init__(self, getter, setter, num_values, check_item, check_active, set_active):
        self._getter = getter
        self._setter = setter
        self._num_values = num_values
        self._check_item = check_item
        self._check_active = check_active
        self._set_active = set_active

    def get_values(self, other):
        """
        Import parameter values.
        Get all values from *other*, which must be another
        :class:`.ParamMatrix` or a nested sequence (such as a
        :class:`list` or :class:`list`). The dimension of *other* must
        be the same as the current instance. Not that in the input
        object is not a :class:`.ParamMatrix`, any value it can have on
        the diagonal will be ignored (but they must be present to avoid
        shifting indexes).
        """

        if isinstance(other, ParamMatrix):
            if other._num_values != self._num_values: raise ValueError('`other` must have the same dimension')
            if not other._check_active():
                if self._check_active(): self._set_active(False)
                return
            if not self._check_active():
                self._set_active(True)
            for i in range(self._num_values):
                for j in range(self._num_values):
                    if i != j: self._setter(i, j, other._getter(i, j))
                        # use of other._getter requires that number has been checked
        else:
            if not self._check_active(): self._set_active(True)
            if len(other) != self._num_values: raise ValueError('`other` must have the same dimension')
            for i, row in enumerate(other):
                if len(row) != self._num_values: raise ValueError('`other` must have the same dimension')
                for j, value in enumerate(row):
                    if i != j:
                        if self._check_item(value) == False: raise ValueError('value out of range')
                        self._setter(i, j, value)

    def __getitem__(self, idx):
        if len(idx) != 2: raise ValueError('expect a tuple of two values')
        if idx[0] < 0: idx[0] = self._num_values + idx[0]
        if idx[1] < 0: idx[1] = self._num_values + idx[1]
        if idx[0] >= self._num_values: raise ValueError('matrix index out of range')
        if idx[1] >= self._num_values: raise ValueError('matrix index out of range')
        if idx[0] == idx[1]: return None
        return self._getter(idx[0], idx[1])

    def __setitem__(self, idx, value):
        if len(idx) != 2: raise ValueError('expect a tuple of two values')
        if idx[0] < 0: idx[0] = self._num_values + idx[0]
        if idx[1] < 0: idx[1] = self._num_values + idx[1]
        if idx[0] >= self._num_values: raise ValueError('matrix index out of range')
        if idx[1] >= self._num_values: raise ValueError('matrix index out of range')
        if idx[0] == idx[1]:
            if value is not None: raise ValueError('cannot set matrix diagonal to a value different than `None`')
            else: return
        if self._check_item(value) == False: raise ValueError('value out of range')
        if not self._check_active(): self._set_active(True)
        self._setter(idx[0], idx[1], value)

    def __iter__(self):
        for i in range(self._num_values):
            for j in range(self._num_values):
                if i == j: yield None
                else: yield self._getter(i, j)

    def count(self, value):
        """
        Number of items that are equal to *value*.
        """
        cnt = 0
        for i in range(self._num_values):
            for j in range(self._num_values):
                if i!=j and self._getter(i, j) == value: cnt += 1
        return cnt

    def index(self, value, imin=0, imax=None, jmin=0, jmax=None):
        """
        Get the index of a given value.
        Return the ``(i, j)`` tuple of indexes of the first value
        (iterating first over rows and then over columns) matching
        *value*. The returned row index is >=imin and <imax and the
        returns column index is >=jmin and <jmax if they are provided.
        Raise a :exc:`ValueError` if the value is not found.
        The diagonal is not considered.
        """
        if imax is None: imax = self._num_values
        if jmax is None: jmax = self._num_values
        for i in range(imin, imax):
            for j in range(jmin, jmax):
                if i!=j and self._getter(i, j) == value: return (i, j)
        else:
            raise ValueError('{0} is not in matrix'.format(value))

    def __contains__(self, value):
        for i in range(self._num_values):
            for j in range(self._num_values):
                if i!=j and self._getter(i, j) == value: return True
        else:
            return False

    def __len__(self):
        return self._num_values

    def __str__(self):
        return str([[self[i,j] for j in range(self._num_values)] for i in range(self._num_values)])

    def __repr__(self):
        if self._num_values <= 4: return str([[self[i,j] for j in range(self._num_values)] for i in range(self._num_values)])
        else: return '<matrix of {0}*{0} values>'.format(self._num_values)

    def __reversed__(self):
        for i in reversed(range(self._num_values)):
            for j in reversed(range(self._num_values)):
                if i == j: yield None
                else: yield self._getter(i, j)

class EventList(object):
    """
    Class storing the list of demographic events. Even if events appear
    as unsorted, they are internally sorted so that the user does not
    have to care about their order. This class has limited
    functionality: add events (but not removing them), accessing and
    modifying parameters.

    In addition to the methods documented below, :class:`.EventList`
    instances support the following operations (where ``events`` is an
    :class:`.EventList` instance):

    +--------------------------+----------------------------------------------------------------------+
    |Expression                | Action                                                               |
    +==========================+======================================================================+
    | ``len(events)``          | Number of events currently loaded                                    |
    +--------------------------+----------------------------------------------------------------------+
    | ``events[i]``            | Dctionary with parameters of the *i*\ th event (see note)            |
    +--------------------------+----------------------------------------------------------------------+
    | ``for event in events``  | Same as ``for i in range(len(events)): event = events[i]``           |
    +--------------------------+----------------------------------------------------------------------+
    | ``str(events)```         | Representation of the instance, roughly as a list                    |
    +--------------------------+----------------------------------------------------------------------+

    The string representation is such as a string at the first level,
    but each event is represented as an angle-bracketed delimited string
    containing comma-separated ``key=value`` pairs with the parameter as
    ``key`` and its value  as ``value`` (with an additional key,
    ``event_index``, providing the index of the event in the list).

    .. note::

        There is no way to modify content of the instance using the
        indexing operator ``events[i]``. One must use
        :meth:`~.EventList.update`.

    More information on the list of events and their parameters is
    available in the documentation for class :class:`.Simulator`.
    """

    _map_enum = {
        'size':          _eggwrapper.Event.change_N,
        'migr':          _eggwrapper.Event.change_M,
        'pair_migr':     _eggwrapper.Event.change_Mp,
        'growth':        _eggwrapper.Event.change_G,
        'selfing':       _eggwrapper.Event.change_s,
        'recombination': _eggwrapper.Event.change_R,
        'bottleneck':    _eggwrapper.Event.bottleneck,
        'admixture':     _eggwrapper.Event.admixture,
        'sample':        _eggwrapper.Event.delayed,
        'merge':         None
    }

    _required_parameters = {  # not: T is enforced in add's signature
        'size':          set(['N']),
        'migr':          set(['M']),
        'pair_migr':     set(['src', 'dst', 'M']),
        'growth':        set(['G']),
        'selfing':       set(['s']),
        'recombination': set(['R']),
        'bottleneck':    set(['S']),
        'admixture':     set(['src', 'dst', 'proba']),
        'sample':        set(['idx', 'label', 'num_chrom', 'num_indiv']),
        'merge':         set(['src', 'dst'])
    }

    _optional_parameters = {
        'size':       set(['idx']),
        'growth':     set(['idx']),
        'selfing':    set(['idx']),
        'bottleneck': set(['idx'])
    }

    _all_parameters = {}
    for i in _required_parameters:
        _all_parameters[i] = _required_parameters[i] | _optional_parameters.get(i, set())
        _all_parameters[i].add('T')

    def __init__(self, npop, add, clear):
        self._npop = npop
        self._add = add
        self._clear = clear
        self._events = []

    def replace(self, other):
        """
        Replace list of events.
        Replace own list of events with the one in the
        :class:`.EventList` instance passed as *other*. The current list
        of events is dropped.
        """
        self.clear() # let self.clear() call self._clear() for security
        for event in other: self.add(** event)

    def clear(self):
        """
        Clear list of events.
        """
        self._clear() # do this first or the Event objects might be garbage-collected!
        del self._events[:]

    def __len__(self):
        return len(self._events)

    def __iter__(self):
        for params, changes in self._events:
            yield dict(params)

    def __str__(self):
        return ('[' + ', '.join([
                '<event_index={0};'.format(i)
                    + ';'.join(['{0}={1}'.format(k,v) for (k,v) in d.items()]) + '>'
                        for i, (d, changes) in enumerate(self._events)]) + ']')

    def __repr__(self):
        return str(self)

    def __getitem__(self, i):
        if i >= len(self._events): raise ValueError('invalid event index')
        return dict(self._events[i][0])

    def add(self, cat, T, **params):
        """
        Add an event to the list. See the documentation for the class
        :class:`.Simulator` for more details on what is expected as
        arguments.

        :param cat: category of the event.
        :param T: date of the event.
        :param params: all needed parameters for this category of
            events.
        """

        # check that category is valid
        if cat not in self._all_parameters:
            raise ValueError('invalid event category: `{0}`'.format(cat))

        # check that all required parameters are present
        if not self._required_parameters[cat].issubset(params):
            raise ValueError('event `{0}` requires: {1}'.format(cat, ', '.join(self._required_parameters[cat] - set(params.keys()))))

        # create the required backend objects
        if cat == 'merge':
            changes = [_eggwrapper.Event(self._map_enum['admixture'], T)]
            if (params['src'] < 0 or params['src'] >= self._npop or
                params['dst'] < 0 or params['dst'] >= self._npop or
                params['src'] == params['dst']):
                    raise ValueError('invalid population indexes provided for `merge` event')
            changes[0].set_index(params['src'])
            changes[0].set_dest(params['dst'])
            changes[0].set_param(1.0)
            for i in range(self._npop):
                if i != params['src']:
                    changes.append(_eggwrapper.Event(self._map_enum['pair_migr'], T))
                    changes[-1].set_param(0.0)
                    changes[-1].set_index(i)
                    changes[-1].set_dest(params['src'])
        else:
            changes = [_eggwrapper.Event(self._map_enum[cat], T)]

        # add the event to the internal list
        params['cat'] = cat
        params['T'] = T
        self._events.append((params, changes))
        params_update = dict(params)

        # use the update method to set parameters (delete the event in case of an error)
        del params_update['cat']
        del params_update['T']
        try:
            self.update(len(self._events) - 1, **params_update)
        except ValueError:
            del self._events[-1]
            raise

        # actually add changes to the lower-level Params
        for change in changes: self._add(change)

    def update(self, event_index, **params):
        """
        Modify any parameter from one of the event of the list. If an
        event's date is modified, sorting will be updated automatically.

        :param event_index: index of the event to modify (based on the
            order in which events have been specified with :meth:`.add`,
            which is the same order as events appear when representing
            the instance or iterating).
        :param params: keyword arguments specifying what parameters to
            modify. Only parameters that have to be changed should be
            specified.
        """

        # check index, get category
        if event_index < 0 or event_index >= len(self._events): raise IndexError('invalid event index')
        cat = self._events[event_index][0]['cat']
        change0 = self._events[event_index][1][0] # not used if complex change

        # initialize bool to perform ad-hoc tests
        test1 = False
        test2 = False

        # process complex events
        if cat == 'merge':
            if 'T' in params:
                for change in self._events[event_index][1]:
                    change.move(params['T'])
            if 'src' in params:
                test2 = True
                if params['src'] < 0 or params['src'] > self._npop: raise ValueError('event `{0}`, population index out of range'.format(cat))
                self._events[event_index][1][0].set_index(params['src'])
                for change in self._events[event_index][1][1:]:
                    change.set_dest(params['src'])

            if 'dst' in params:
                test2 = True
                if params['dst'] < 0 or params['dst'] > self._npop: raise ValueError('event `{0}`, population index out of range'.format(cat))
                self._events[event_index][1][0].set_dest(params['dst'])

        # process simple events
        else:

            # process all parameters and check+set them
            for key, value in params.items():
                if key not in self._all_parameters[cat]:
                    raise ValueError('event `{0}`: unknown parameter: {1}'.format(cat, key))

                if key == 'T':
                    if value < 0.0: raise ValueError('event `{0}`: date must be positive'.format(cat))
                    change0.move(value)
                elif key == 'N':
                    if value <= 0.0: raise ValueError('event `{0}`: size must be strictly positive'.format(cat))
                    change0.set_param(value)
                elif key == 'M':
                    if value < 0.0: raise ValueError('event `{0}`: size cannot be negative'.format(cat))
                    change0.set_param(value)
                elif key == 'G':
                    change0.set_param(value)
                elif key == 's':
                    if value < 0.0 or value > 1.0: raise ValueError('event `{0}`: selfing rate must be between 0 and 1'.format(cat))
                    change0.set_param(value)
                elif key == 'R':
                    if value < 0.0: raise ValueError('event `{0}`: recombination rate must be positive'.format(cat))
                    change0.set_param(value)
                elif key == 'S':
                    if value < 0.0: raise ValueError('event `{0}`: size must be positive'.format(cat))
                    change0.set_param(value)
                elif key == 'proba':
                    if value < 0.0 or value > 1.0: raise ValueError('event `{0}`: probability must be between 0 and 1'.format(cat))
                    change0.set_param(value)
                elif key == 'idx':
                    if value < 0 or value > self._npop: raise ValueError('event `{0}`: population index out of range'.format(cat))
                    change0.set_index(value)
                elif key == 'src':
                    test2 = True
                    if value < 0 or value > self._npop: raise ValueError('event `{0}`: population index out of range'.format(cat))
                    change0.set_index(value)
                elif key == 'dst':
                    test2 = True
                    if value < 0 or value > self._npop: raise ValueError('event `{0}`: population index out of range'.format(cat))
                    change0.set_dest(value)
                elif key == 'label':
                    if not isinstance(value, str): raise TypeError('label must be a string')
                    change0.set_label(value)
                elif key == 'num_chrom':
                    test1 = True
                    if value < 0: raise ValueError('event `{0}`: number of samples must be positive'.format(cat))
                    change0.set_number1(value)
                elif key == 'num_indiv':
                    test1 = True
                    if value < 0: raise ValueError('event `{0}`: number of samples must be positive'.format(cat))
                    change0.set_number2(value)
                else:
                    raise RuntimeError('unexpected error')

        # perform ad-hoc tests
        if test1 and change0.get_number1() + change0.get_number2() == 0:
            raise ValueError('event `{0}`: there must be at least one sample'.format(cat))

        if test2 and change0.get_index() == change0.get_dest():
            raise ValueError('event `{0}`: cannot source/destination populations cannot be the same'.format(cat))

        # copy all parameters
        self._events[event_index][0].update(params)



"""
    Copyright 2009-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import subprocess, re, os
from .. import _interface, _tree, alphabets
from ..tools import _code_tools
from . import _utils

_codeml_ctl_template = """
      seqfile = i
     treefile = t
      outfile = o
        noisy = 3
      verbose = 0
      runmode = 0
      seqtype = 1
    CodonFreq = {codon_freq}
        clock = 0
       aaDist = 0
        model = {model}
      NSsites = {ns_sites}
        icode = 0
        Mgene = 0
    fix_kappa = {fix_kappa}
        kappa = {kappa}
    fix_omega = {fix_omega}
        omega = {omega}
    fix_alpha = 1
        alpha = 0.0
       Malpha = 0
        ncatG = {ncatg}
        getSE = 0
 RateAncestor = 1
   Small_Diff = .5e-6
    cleandata = 0
       method = 0
"""

class _Codeml(_utils._App):

    @_utils._protect_run
    def _check_path(self, path, cfg):

        # test the "help" option to ensure that a phyml exist
        f = open('a', 'w')
        f.write('\n')
        f.close()
        cmd = (path, 'a')
        try:
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
            stdout, stderr = p.communicate()
            mo = re.search('version (\d)\.(\d+)([a-z]?)', stdout)
            if not mo: return 'cannot read version number in CodeML output'
            M = int(mo.group(1))
            m = int(mo.group(2))
            r = mo.group(3)
        except OSError as e:
            return e.strerror

        if M != 4 or m < 8: return 'PAML version 4.8 or higher is required'
        return None

_app = _Codeml(key='codeml', default='codeml')
_utils.paths._add(_app)

_models = {

    # columns of the models table
        # description
        # value for `model` in codeml.ctl file
        # value for `NSsites` in codeml.ctl file
        # omega value fixed by model
        # required number of catgories
        # site omega: 0=none, 1=NEB, 2=BEB
        # dN and dS tree lengths
        # discrete ratios:
        #       0 (one ratio)
        #       1 (discrete)
        #       2 (branch-site A)
        #       3 (branch-site C/D)
        #       7 (beta models M7, M8)
        # require tags: 0=no, 1=yes, 2=exactly two
    # key    description             m  NS  omega ncat   w/s lens   dis tags
    'M0':   ('one ratio',            0, 0,  None, False, 0,  True,  0,  0),
    'free': ('one ratio per branch', 1, 0,  None, False, 0,  True,  0,  0),
    'nW':   ('sets of branches',     2, 0,  None, False, 0,  True,  0,  1),
    'M1a':  ('nearly neutral',       0, 1,  None, False, 1,  False, 1,  0),
    'M2a':  ('positive selection',   0, 2,  None, False, 2,  False, 1,  0),
    'M3':   ('discrete',             0, 3,  None, True,  1,  False, 1,  0),
    'M4':   ('freqs',                0, 4,  None, True,  1,  False, 1,  0),
    'M7':   ('beta',                 0, 7,  None, True,  1,  False, 7,  0),
    'M8a':  ('beta + omega=1',       0, 8,  1,    True,  1,  False, 7,  0),
    'M8':   ('beta + omega>1',       0, 8,  None, True,  2,  False, 7,  0),
    'A0':   ('null model for A',     2, 2,  1,    False, 1,  False, 2,  2),
    'A':    ('branch-site model',    2, 2,  None, False, 2,  False, 2,  2),
    'C0':   ('null model for C',     0, 22, None, False, 1,  False, 1,  0),
    'C':    ('clade model C',        3, 2,  None, False, 2,  False, 3,  1),
    'D':    ('clade model D',        3, 3,  None, True,  1,  False, 3,  1)
}

@_utils._protect_run
def codeml(align, tree, model, code=1, ncat=None, codon_freq=2,
    verbose=False, get_files=False, kappa=2.0,
    fix_kappa=False, omega=0.4):

    """ codeml(align, tree, model, code=1, ncat=None, codon_freq=2, \
               verbose=False, get_files=False, \
               kappa=2.0, fix_kappa=False, omega=0.4)

    Fit nucleotide substitution models using `PAML <http://abacus.gene.ucl.ac.uk/software/paml.html>`_.

    This function uses the CodeML program only of the PAML package.

    :param align: an :class:`.Align` containing a coding sequence
        alignment. The number of sequences must be at least 3, the
        length of the alignment is required to be a multiple of 3
        (unless codons are provided). There must be no stop codons (even
        final stop codons) and there must not be any duplicated sequence
        name. The alphabet might be DNA of codon.

    :param tree: a :class:`.Tree` providing the phylogenetic
        relationships between samples. The name of the sequences in the
        :class:`.Align` and in the :class:`.Tree` are required to match.
        If *tree* is ``None``, a star topology is used. If the tree
        contains branch length or node labels, they are discounted,
        except for PAML node tags (``#x`` and ``$x`` where ``x`` is an
        integer) that are allowed both as nodel labels. If one wants to
        label a terminal branch of the tree, they can add the label at
        the end of the sample name (with an optional separating white
        space). The tree must not be rooted (if there is a birfurcation
        at the base, an error will be caused).

    :param model: model. The list of model names appears below:

        * ``M0`` -- one-ratio model (1 parameter).
        * ``free`` -- all branches have a different ratio (1 parameter
          per branch).
        * ``nW`` -- several sets of branches. Requires labelling of
          branches of the tree (1 parameter per set of branches).
        * ``M1a`` -- nearly-neutral model (2 parameters).
        * ``M2a`` -- positive selection model (4 parameters).
        * ``M3`` -- discrete model. Requires setting *ncat*  (2 * *ncat*
          - 1 parameters).
        * ``M4`` -- frequencies model. Requires setting *ncat* (*ncat* -
          1 parameters).
        * ``M7`` -- beta-distribution model. Requires setting *ncat* (2
          parameters).
        * ``M8a`` -- beta + single ratio, additional ratio fixed to 1.
          Requires setting *ncat* (3 parameters).
        * ``M8`` -- beta + single ratio. Requires setting *ncat* (4
          parameters).
        * ``A0`` -- null branch-site model. Requires labelling of
          branches of the tree with two different labels (3 parameters).
        * ``A`` -- branch-site model with positive selection. Requires
          labelling of branches of the tree with two different labels (4
          parameters).
        * ``C0`` -- null model for model C (M2a_rel). Does not require
          branch labelling (4 parameters).
        * ``C`` -- branch-site model. Requires labelling of branches (5
          parameters).
        * ``D`` -- discrete branch-site model. Requires labelling of
          branches and requires setting *ncat* to either 2 or 3 (4 or
          6 parameters, respectively).

        The number of parameters given for each model concern the
        *dN/dS* ratios only. Refer to PAML documentation or the
        following references for more details and recommendations:
        Bielawski, J.P. & Z. Yang. 2004. A maximum likelihood method for
        detecting functional divergence at individual codon sites, with
        application to gene family evolution. *J. Mol. Evol.*
        **59**\ :121-132; Yang Z., R. Nielsen, N. Goldman & A.M.K.
        Pedersen. 2000; Codon-substitution models for heterogeneous
        selection pressure at amino acid sites. *Genetics*
        **155**\ :431-449. Yang, Z., and R. Nielsen. 2002.
        Codon-substitution models for detecting molecular adaptation at
        individual sites along specific lineages. *Mol. Biol. Evol.*
        **19**\ :908-917; Zhang, J., R. Nielsen & Z. Yang. 2005.
        Evaluation of an improved branch-site lieklihood method for
        detecting positive selection at the molecular level. *Mol. Biol.
        Evol.* **22**\ :472-2479.

    :param code: genetic code identifier (see :ref:`here <genetic-codes>`).
        Required to be an integer among the valid values. The default
        value is the standard genetic code. Only codes 1-11 are
        available.

    :param ncat: number of *dN/dS* categories. Only a subset of models
        require that the number of categories to be specified. See
        *models*.

    :param codon_freq: an integer specifying the model for codon
        frequencies. Must be one of:

        * 0 -- equal frequencies.
        * 1 -- codon frequencies from base frequencies (3 degrees of
          liberty).
        * 2 -- codon frequencies from base frequencies at three codon
          positions (9 degrees of liberty).
        * 3 -- independent codon frequencies (60 degrees of freedom).

    :param verbose: boolean indicating whether standard output of CodeML
        should be displayed.

    :param get_files: boolean indicating whether the raw content of
        CodeML output files should be included in the returned data.

    :param kappa: starting value for the transition/transversion
        rate ratio.

    :param fix_kappa: tell if the transition/transversion rate ratio
        should be fixed to its starting value (otherwise, it is
        estimated as a free parameter).

    :param omega: starting value for the *dN/dS* ratio (strictly
        positive value).

    :return: A :class:`dict` holding results. The keys defined in the
        returned dictionary are:

        * ``model`` -- model name.
        * ``lk`` -- log-likelihood.
        * ``np`` -- number of parameters of the model.
        * ``kappa`` -- fixed or estimated value of the
          transition/transversion rate ratio.
        * ``beta`` -- if model is ``M7``, ``M8a``, or ``M8``, a
          :class:`tuple` with the *p* and *q* parameters of the beta
          distribution of neutral dN/dS ratios; otherwise, ``None``.
        * ``K`` -- number of dN/dS ratio categories. Equals to 0 for the
          ``free`` model, to the number of branch categories for the
          ``nW`` model, and to the number of site categories otherwise.
          This value is not necessarily equal to the *ncat* argument
          because ``M8a`` and ``M8`` models add a category, and because
          it has a different meaning for model ``nW``.
        * ``num_tags`` -- number of branch categories detected from the
          imported tree (irrespective to the model that has been
          fitted). If the star topology has been used (``tree=None``),
          this value is 1.
        * ``omega`` -- estimated dN/dS ratio or ratios. The structure of
          the value depends on the model:

          * ``M0`` model -- a single value.
          * ``free`` model -- ``None`` (ratios are available as node
            labels in the tree available as ``tree_ratios``).
          * ``nW`` model -- a :class:`list` of dN/dS ratios for all
            branch categories (they are listed in the order
            corresponding to branch labels).
          * Discrete models (``M1a``, ``M2a``, ``M3``, ``M4``, ``C0``,
            ``M7``, ``M8a``, and ``M8``) -- a :class:`!list` of ``K``
            dN/dS ratios. The frequency of each dN/dS category is
            available is ``freq``.
          * ``A0`` and ``A`` models -- a :class:`!tuple` of two
            :class:`!list` of 4 items each, containing respectively the
            background and foreground dN/dS ratios. The frequency of
            each dN/dS category is available is ``freq``.
          * ``C`` and ``D`` models -- a :class:`!tuple` of ``num_tags``
            :class:`!list` (one :class:`!list` for each set of branches,
            as defined by branch labels found in the provided tree),
            each of them containing ``K`` dN/dS ratios.  The frequency
            of each dN/dS category is available is ``freq``.

        * ``freq`` -- the frequency of dN/dS ratio categories. If
          defined, it is a :class:`!list` of ``K`` values. This entry is
          ``None`` for models ``M0``, ``free``, and ``nW``.
        * ``length`` -- total length of tree after estimating branch
          lengths with the specified model.
        * ``tree`` -- the tree with fitted branch lengths, as a
          :class:`.Tree` instance. Branch lengths are expressed in terms
          of the model of codon evolution.
        * ``length_dS`` -- total length of tree in terms of synonymous
          substitutions. Only available with ``M0``, ``free``, and
          ``nW`` models.
        * ``length_dN`` -- total length of tree in terms of
          non-synonymous substitutions. Only available with ``M0``,
          ``free``, and ``nW`` models.
        * ``tree_dS`` -- a :class:`!Tree` instance with branch lengths
          expressed in terms of synonymous substitutions. Only available
          with ``free`` and ``nW`` models.
        * ``tree_dN`` -- a :class:`!Tree` instance with branch lengths
          expressed in terms of non-synonymous substitutions. Only
          available with ``free`` and ``nW`` models.
        * ``tree_ratios`` -- a :class:`!Tree` instance with the dN/dS
          ratios included as branch labels. Only available with ``free``
          and ``nW`` models.
        * ``site_w`` -- a :class:`!dict` containing posterior predictions
          of site dN/dS ratios. Not available for models ``M0``,
          ``free``, and ``nW`` (in that cases, the value is ``None``).
          The :class:`!dict` contains the following keys:

          * ``method`` -- on the strings ``NEB`` and ``BEB``.
          * ``aminoacid`` -- the list of reference amino acids for all
            amino acid sites of the alignment (they are taken from the
            first sequence in the original alignment).
          * ``proba`` -- the list of posterior probabilites of the dN/dS
            categories for all amino acid sites of the alignment. For
            each site, a :class:`!tuple` of ``K`` (the number of dN/dS
            categories) is provided.
          * ``best`` -- the index of the best category for each site.
          * ``postw`` -- list of the posterior dN/dS estimate for all
            sites (``None`` if not available).
          * ``postwsd`` -- list of the standard deviation of the dN/dS
            estimate for all sites (always available if ``postw`` is
            available and the method is ``BEB``, ``None`` otherwise).
          * ``P(w>1)`` -- probability that the dN/dS ratio is greater
            than 1 for all sites (``None`` if not available).

        * ``main_output`` -- raw content of the main CodeML output file.
          This key is not present if the option *get_files* is not set
          to ``True``.
        * ``rst_output`` -- raw content of the ``rst`` detailed CodeML
          output file. This key is not present if the option *get_files*
          is not set to ``True``.

    .. versionchanged:: 3.0.0

        Turned into a singe function, interface changes (more models,
        more options, more results).
    """

    # check that program is available
    path = _app.get_path()
    if path is None:
        raise RuntimeError('PAML\'s codeml program not available -- please configure path')

    # process input alignment
    if not isinstance(align, _interface.Align): raise TypeError('`align` must be an Align instance')
    if len(align) < 3: raise ValueError('not enough sequences in alignment')
    if align._alphabet._obj.get_type() == 'DNA':
        if align.ls < 3 or align.ls % 3 != 0: raise ValueError('alignment length must be a multiple of 3')
    elif align._alphabet._obj.get_type() != 'codons':
        raise ValueError('invalid alphabet for coding sequences')

    # get code
    if code not in _code_tools._codes: raise ValueError('unknown genetic code: {0}'.format(code))
    if code > 11: raise ValueError('unsupported genetic code: {0}'.format(code))
    _code = _code_tools._codes[code]

    # check for stop codons
    if _code_tools.has_stop(align, code=code): raise ValueError('stop codon found in sequences')

    # write down alignment
    mapping = {}
    _utils._write(align, 'i', mapping)

    # revert mapping
    rmapping = {}
    for key, sample in mapping.items():
        if sample.name == '':
            raise ValueError('alignment contains an empty-string name')
        if sample.name in rmapping:
            raise ValueError('alignment contains duplicates: {0}'.format(sample.name))
        rmapping[sample.name] = key

    # check and print tree
    tags = set()
    if tree == None:
        f = open('t', 'w')
        f.write('(' + ','.join(mapping) + ');\n')
        f.close()
        tags.add(0)
    else:
        tree = tree.copy() # make copy to allow modify it
        if tree.base.num_children < 3: raise ValueError('tree must not be rooted')
        if tree.num_leaves != len(rmapping): raise ValueError('number of sequences does not match with tree')
        for node in tree.depth_iter():
            # if terminal node, allow label to be: "name[[]#|$tag]"
            if node.num_children == 0:
                mo = re.match('(.+?)(?: ?([\#|\$])(\d+))?$', node.label)
                if mo is None: raise ValueError('invalid name in tree: {0}'.format(node.label))
                name, symb, tag = mo.groups()
                if name not in rmapping: raise ValueError('name in tree does not match with sequence names: {0}'.format(leaf.label))
                if tag is None: node.label = rmapping[name]
                else: node.label = rmapping[name] + ' ' + symb + tag

            # if internal node, allow None, an integer (that it deleted), or #|$tag
            elif node.num_children > 1:
                tag = None
                if node.label is None: pass
                elif isinstance(node.label, int): node.label = None
                else:
                    mo = re.match('[\#|\$](\d+)$', node.label)
                    if mo is None: raise ValueError('invalid internal node label: {0}'.format(node.label))
                    tag = mo.group(1)

            # safety check
            else: raise ValueError('invalid tree structure')

            # manage tag
            if tag is None: tags.add(0)
            else: tags.add(int(tag))

        if sorted(tags) != list(range(len(tags))): raise ValueError('invalid tree labels (labels must start from 0 and must be consecutive)')
        f = open('t', 'w')
        f.write(tree.newick(skip_brlens=True) + '\n')
        f.close()

    # check model
    if model not in _models: raise ValueError('invalid model name: {0}'.format(model))
    doc, mod, ns, w, req_ncat, has_beb, has_treelen, discrete, req_tags = _models[model]
    if model == 'D' and ncat != 2 and ncat != 3: raise ValueError('`ncat` must be set to 2 or 3 when used with model D')
    if req_ncat:
        if ncat is None: raise ValueError('required `ncat` not specified')
        if not isinstance(ncat, int): raise TypeError('`ncat` must be an integer')
        if ncat < 2: raise ValueError('`ncat` must be at least 1')
    else:
        ncat = 0 # should be ignored
    if req_tags > 0 and len(tags) < 2: raise ValueError('model {0} requires at least two different tags in tree'.format(model))
    if req_tags == 2 and len(tags) != 2: raise ValueError('model {0} requires exactly two different tags in tree'.format(model))

    # check other arguments
    if codon_freq not in list(range(4)): raise ValueError('invalid value for `codon_freq` argument')
    if kappa <= 0.0: raise ValueError('invalid value for `kappa` argument')
    if omega <= 0.0 or omega >= 89.0: raise ValueError('invalid value for `omega` argument')

    # create control file
    if w == None:
        w = omega
        fix_omega = 0
    else:
        fix_omega = 1

    ctl =  _codeml_ctl_template.format(codon_freq=codon_freq,
                                       model=mod,
                                       ns_sites=ns,
                                       fix_omega=fix_omega,
                                       omega=w,
                                       ncatg=ncat,
                                       fix_kappa=1 if fix_kappa else 0,
                                       kappa=kappa)

    f = open('codeml.ctl', 'w')
    f.write(ctl)
    f.close()

    # run codeml
    p = subprocess.Popen((path, 'codeml.ctl'), stdin=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        stdout=(None if verbose else subprocess.PIPE), universal_newlines=True)
    stdout, stderr = p.communicate()

    # check error
    stderr = stderr.strip()
    if len(stderr):
        raise RuntimeError('error while running codeml: {0}'.format(stderr))
    if not os.path.isfile('rst') or not os.path.isfile('o'):
        raise RuntimeError('unknown error while running codeml (try running in verbose mode)')

    # get data from main output file
    final = {'model': model, 'num_tags': len(tags)}
    res = _helper_main(locals())
    if isinstance(res, int): raise ValueError('cannot read results from codeml output file (try running in verbose mode) [code: {0}]'.format(res))
    final.update(res)

    # get data from rst file
    final['site_w'] = _helper_rst(locals())

    # return
    return final

def _helper_main(variables):

    res = {}

    f = open('o')
    main = f.read()
    f.close()
    if variables['get_files']: res['main_output'] = main

    # lnL & np
    mo = re.search('lnL\(ntime: ?.+  np: ?(.+)\): *(.+) +\+.+', main)
    if mo is None: return 1
    np, lnL = mo.groups()
    res['np'] = int(np)
    res['lnL'] = float(lnL)

    # kappa
    if not variables['fix_kappa']:
        mo = re.search('kappa \(ts/tv\) = +(.+)', main)
        if mo is None: return 2
        res['kappa'] = float(mo.group(1))
    else:
        mo = re.search('kappa = (.+) fixed', main)
        if mo is None: return 7
        res['kappa'] = float(mo.group(1))

    # dN and dS tree lengths
    if variables['has_treelen']:
        mo = re.search('tree length for dN: +(.+)', main)
        if mo is None: return 3
        res['length_dN'] = float(mo.group(1))

        mo = re.search('tree length for dS: +(.+)', main)
        if mo is None: return 4
        res['length_dS'] = float(mo.group(1))
    else:
        res['length_dN'] = None
        res['length_dS'] = None

    # omega or omega classes
    if variables['model'] == 'free':
        res['omega'] = None
        res['freq'] = None
        res['K'] = 0

    elif variables['model'] == 'nW':
        reg = '^w \(dN/dS\) for branches:' + ' +([\.\dEe]+)' * len(variables['tags']) + '$'
        mo = re.search(reg, main, re.MULTILINE)
        if mo is None: return 12
        res['omega'] = tuple(map(float, mo.groups()))
        res['freq'] = None
        res['K'] = len(res['omega'])

    elif variables['discrete'] == 0:
        mo = re.search('omega \(dN/dS\) = +(.+)', main)
        if mo is None: return 5
        res['omega'] = float(mo.group(1))
        res['freq'] = None
        res['K'] = 1

    else:
        mo = re.search('dN/dS \(w\) for site classes \(K=(\d+)\)', main)
        if mo is None: return 6
        K = int(mo.group(1))
        res['K'] = K

        if variables['discrete'] == 1:
            reg = '^p:' + ''.join([' +([\.\dEe]+)']*K) + '\n' + 'w:' + ''.join([' +([\.\dEe]+)']*K) + '$'
            mo = re.search(reg, main, re.MULTILINE)
            if mo is None: return 7
            grp = mo.groups()
            res['freq'] = tuple(map(float, grp[:K]))
            res['omega'] = tuple(map(float, grp[K:]))

        elif variables['discrete'] == 2:
            if K != 4: return 8
            reg = """^site class +0 +1 +2a +2b
proportion +([\.\dEe]+) +([\.\dEe]+) +([\.\dEe]+) +([\.\dEe]+)
background w +([\.\dEe]+) +([\.\dEe]+) +([\.\dEe]+) +([\.\dEe]+)
foreground w +([\.\dEe]+) +([\.\dEe]+) +([\.\dEe]+) +([\.\dEe]+)$"""
            mo = re.search(reg, main, re.MULTILINE)
            if mo is None: return 9
            grp = mo.groups()
            res['freq'] = tuple(map(float, grp[:4]))
            res['omega'] = tuple(map(float, grp[4:8]), map(float, grp[8:]))

        elif variables['discrete'] == 3:
            reg = []
            reg.append('^site class' + ''.join([' +{0}'.format(i) for i in range(K)]))
            reg.append('proportion' + ' +([\.\dEe]+)' * K)
            for i in range(len(variables['tags'])):
                reg.append('branch type {0}:'.format(i) + ' +([\.\dEe]+)' * K)
            reg = '\n'.join(reg)
            reg += '$'
            mo = re.search(reg, main, re.MULTILINE)
            if mo is None: return 10
            grp = mo.groups()
            res['freq'] = tuple(map(float, grp[:K]))
            res['omega'] = tuple(tuple(map(float, grp[K+i*K:2*K+i*K])) for i in range(len(variables['tags'])))

        elif variables['discrete'] == 7:
            reg = '^p:' + ' +([\.\dEe]+)' * K + '\n' + 'w:' + ' +([\.\dEe]+)' * K + '$'
            mo = re.search(reg, main, re.MULTILINE)
            if mo is None: return 11
            grp = mo.groups()
            res['freq'] = tuple(map(float, grp[:K]))
            res['omega'] = tuple(map(float, grp[K:]))

        else:
            return 99

    # beta distribution parameters
    if variables['model'] == 'M7':
        reg = '^Parameters in M7 \(beta\):\n p = +([\.\dEe]+)  q = +([\.\dEe]+)$'
        mo = re.search(reg, main, re.MULTILINE)
        if mo is None: return 13
        res['beta'] = float(mo.group(1)), float(mo.group(2))
    elif variables['model'] == 'M8a' or variables['model'] == 'M8':
        reg = """^Parameters in M8 \(beta&w>1\):
  p0 = +[\.\dEe]+  p = +([\.\dEe]+) q = +([\.\dEe]+)
 \(p1 = +[\.\dEe]+\) w = +[\.\dEe]+$"""
        mo = re.search(reg, main, re.MULTILINE)
        if mo is None: return 14
        res['beta'] = float(mo.group(1)), float(mo.group(2))
    else:
        res['beta'] = None

    # tee length and tree with branch lengths
    reg = """^tree length = +([\.\dEe]+)

(\([ (),\-\.\d:eqs]+\);)

(\([ (),\-\.\d:eqs]+\);)$"""
    mo = re.search(reg, main, re.MULTILINE)
    if mo is None: return 15
    res['length'] = float(mo.group(1))
    res['tree'] = _tree.Tree(string=mo.group(3))
    for leaf in res['tree'].iter_leaves(): leaf.label = variables['mapping'][leaf.label].name

    # dN/dS and w-ratio-labelled trees
    if variables['model'] == 'nW' or variables['model'] == 'free':

        reg = """^dS tree:
(\([ (),\-\.\d:eqs]+\);)
dN tree:
(\([ (),\-\.\d:eqs]+\);)

w ratios as labels for TreeView:
(\([# (),\-\.\d:eqs]+\);)$"""
        mo = re.search(reg, main, re.MULTILINE)
        if mo is None: return 16
        res['tree_dS'] = _tree.Tree(string=mo.group(1))
        res['tree_dN'] = _tree.Tree(string=mo.group(2))
        for leaf in res['tree_dS'].iter_leaves(): leaf.label = variables['mapping'][leaf.label].name
        for leaf in res['tree_dN'].iter_leaves(): leaf.label = variables['mapping'][leaf.label].name
        res['tree_ratios'] = _tree.Tree(string=mo.group(3))

    else:
        res['tree_dS'] = None
        res['tree_dN'] = None
        res['tree_ratios'] = None

    return res

def _helper_rst(variables):

    f = open('rst')
    rst = f.read()
    f.close()
    if variables['get_files']: res['rst_output'] = main

    if variables['has_beb'] == 0:
        return None

    else:

        # make the whole-block regular expression
        method_full = 'Naive' if variables['has_beb'] == 1 else 'Bayes'
        method = method_full[0] + 'EB'

        reg = ['^{0} Empirical Bayes \({0[0]}EB\) probabilities for (\d+) classes( \(class\))?( & postmean_w)?( & P\(w>1\))?'.format(method_full)]
        reg.append('\(amino acids refer to 1st sequence: seq-1\)')
        reg.append('')
        nsites = variables['align'].ls//3
        for i in range(nsites):
            reg.append(' +{0} +[A-Z].+'.format(i+1))
        reg = '\n'.join(reg)
        reg += '$'

        mo = re.search(reg, rst, re.MULTILINE)
        if not mo: raise ValueError('cannot find {0} results in rst file'.format(method))

        hit = mo.group(0)
        K = int(mo.group(1))
        if method == 'BEB' and mo.group(2) is None: raise ValueError('inconsistency in BEB output')
        post_mean = mo.group(3) is not None
        post_test = mo.group(4) is not None
        if post_test and not post_mean: raise ValueError('unexpected case in {0} output'.format(method))
        if post_test and post_mean and method=='BEB': raise ValueError('unexpected case in BEB output')

        # initialize the result dict
        res = {
            'method': method,
            'aminoacid': [],
            'proba': [],
            'best': [],
            'postw': None,
            'postwsd': None,
            'P(w>1)': None
        }
        if post_mean:
            res['postw'] = []
            if method == 'BEB': res['postwsd'] = []
            if post_test: res['P(w>1)'] = []

        # get the posterior predictions for each site
        templ = '^ +{0} ([A-Z])' + ' +([\.\dEe]+)' * K + ' \( *(\d+)\)'
        if post_mean:
            if method == 'NEB': templ += ' +([\.\dEe]+)'
            else: templ += ' +([\.\dEe]+) \+- +([\.\dEe]+)'
            if post_test: templ += ' +([\.\dEe]+)'
        templ += '$'

        for i in range(nsites):
            mo = re.search(templ.format(i+1), hit, re.MULTILINE)
            if mo is None: raise ValueError('cannot find {0} results for site {1}'.format(method, i+1))
            hits = mo.groups()
            base = hits[0]
            proba = tuple(map(float, hits[1:K+1]))
            best = int(hits[K+1]) - 1
            res['aminoacid'].append(base)
            res['proba'].append(tuple(proba))
            res['best'].append(best)
            if post_mean:
                postw = float(hits[K+2])
                res['postw'].append(postw)
                if method == 'BEB':
                    postwsd = float(hits[K+3])
                    res['postwsd'].append(postwsd)
                    if len(hits) != K+4: raise ValueError('error while reading BEB result for site {0}'.format(i+1))
                elif post_test:
                    P = float(hits[K+3]) # we have tested that BEB must be false
                    res['P(w>1)'].append(P)
                    if len(hits) != K+4: raise ValueError('error while reading {0} result for site {1}'.format(method, i+1))
                else:
                    if len(hits) != K+3: raise ValueError('error while reading {0} result for site {1}'.format(method, i+1))
            elif len(hits) != K+2: raise ValueError('error while reading {0} result for site {1}'.format(method, i+1))

        # return
        return res



"""
    Copyright 2009-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import os, re, subprocess
from .. import _interface, _tree, alphabets
from ._utils import _App, _protect_run, paths

_example_file = """ 4 47
one       ACGGCGCTATAAATCGAGCGTTAGCGCGCGGAGAGAGACCTCTCTAG
two       AGGGCGCTATATATCGCGCGTTTGCGCGCGGAGAGAGAGCTCTCTAG
three     AGGGCGGTTTATATCGGGCGTTTTCGCCCGGTCCGAGAGCACTCTAG
four      ACGGCGCTTTATATCCGCCGTTAACGCCCGGTCCGAGGGCACTCTAG
"""

class _Phyml(_App):

    @_protect_run
    def _check_path(self, path, cfg):

        # test the "help" option to ensure that a phyml exist
        cmd = (path, '--help')
        try:
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = p.communicate()
            if len(stderr): return stderr
        except OSError as e:
            return e.strerror

        # test that phyml works
        f = open('input.phy', 'w')
        f.write(_example_file)
        f.close()

        cmd = (path, '-i', 'input.phy', '-d', 'nt', '-b', '10',
               '-m', 'HKY85', '-c', '4', '-a', 'e')

        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        stdout, stderr = p.communicate()
        if len(stderr): return stderr
        if p.returncode != 0:
            return f'PhyML version is probably wrong (version 3.1 or higher is required)\nreturn code: {p.returncode}\nstderr: {stderr}'

        if not os.path.isfile('input.phy_phyml_tree')       and not os.path.isfile('input.phy_phyml_tree.txt'):       return 'output file not generated'
        if not os.path.isfile('input.phy_phyml_stats')      and not os.path.isfile('input.phy_phyml_stats.txt'):      return 'output file not generated'
        if not os.path.isfile('input.phy_phyml_boot_stats') and not os.path.isfile('input.phy_phyml_boot_stats.txt'): return 'output file not generated'
        if not os.path.isfile('input.phy_phyml_boot_trees') and not os.path.isfile('input.phy_phyml_boot_trees.txt'): return 'output file not generated'

        # check the version
        cmd = (path, '--version')
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        stdout, stderr = p.communicate()
        if len(stderr): return stderr
        mo = re.search('This is PhyML version 3\.(\d+)\.(\d+)', stdout)
        if mo is None:
            mo = re.search('This is PhyML version (\d+)', stdout)
            if mo is None:
                return 'cannot check version (invalid version string)'
            version = int(mo.group(1))
            if version < 20120000: return 'PhyML version 3.1 or higher is required'

        return None

_app = _Phyml(key='phyml', default='phyml')
paths._add(_app)

_nt_models = ['JC69', 'K80', 'F81', 'HKY85', 'F84', 'TN93', 'GTR']
_aa_models = ['LG', 'WAG', 'JTT', 'MtREV', 'Dayhoff', 'DCMut', 'RtREV',
              'CpREV', 'VT', 'Blosum62', 'MtMam', 'MtArt', 'HIVw',
              'HIVb']

@_protect_run
def phyml(align, model, labels=False,
           rates=1, boot=0, start_tree='nj', fixed_topology=False,
           fixed_brlens=False, freq=None, TiTv=4.0, pinv=0.0, alpha=None, use_median=False,
           free_rates=False, seed=None, verbose=False):

    """ phyml(align, model, labels=False,
           rates=1, boot=0, start_tree='nj', fixed_topology=False,
           fixed_brlens=False, freq=None, TiTv=4.0, pinv=0.0, alpha=None, use_median=False,
           free_rates=False, seed=None, verbose=False)

    Reconstruct maximum-likelihood phylogeny using `PhyML <http://www.atgc-montpellier.fr/phyml/>`_.

    .. include:: <isoamsa.txt>

    PhyML is a program performing maximum-likelihood phylogeny
    estimation using nucleotide or amino acid sequence alignments.

    .. admonition:: Reference

        Guindon S., J.-F. Dufayard, V. Lefort, M. Animisova, W.
        Hordijk, and O. Gascuel. 2010. New algorithms and methods to
        estimate maximum-likelihood phylogenies: assessing the performance
        of PhyML 3.0. *Syst. Biol.* **59**\ : 307-321.

    :param align: input sequence alignment as an :class:`.Align`
        instance.
    :param model: substitution model to use (see list below).
    :param labels: boolean indicating whether the group labels
        should be included in the names of sequences (they will as the
        following string: `@lbl1,lbl2,lbl3...`, that is: @ followed by
        all labels separated by commas).
    :param rates: number of discrete categories of evolutionary rate. If
        different of 1, fits a gamma distribution of rates.
    :param boot: number of bootstrap repetitions. Values of -1, -2 and
        -4 activate one the test-based branch support evaluation methods
        that provide faster alternatives to bootstrap repetitions (-1:
        aLRT statistics, -2: Chi2-based parametric tests, -4: Shimodaira
        and Hasegawa-like statistics). A value of 0 provides no branch
        support at all.
    :param start_tree: starting topology used by the program. Possible
        values are the string ``nj`` (neighbour-joining tree), the
        string ``pars`` (maximum-parsimony tree), and a
        :class:`.Tree` instance containing a user-provided topology. In
        the latter case, the names of leaves of the tree must match the
        names of the input alignment (without group labels), implying
        that names cannot be repeated.
    :param fixed_topology: boolean indicating whether the topology
        provided in the :class:`.Tree` instance passed as *start_tree*
        argument should be NOT be improved.
    :param fixed_brlens: boolean indicating whether the branch lengths
        provided in the :class:`.Tree` instance passed as *start_tree*
        argument should be NOT be improved. All branch lengths must be
        specified. Automatically sets *fixed_topology* to ``True``.
    :param freq: nucleotide or amino acid frequencies. Possible values
        are the string ``o`` (observed, frequencies measured from the
        data), the string ``m`` (estimated by maximum likelihood for
        nucleotides, or retrieved from the substitution model for amino
        acids), or a four-item tuples provided the relative frequencies
        of A, C, G and T respectively (only for nucleotides). By
        default, use ``o``  for nucleotides and ``m`` for amino acids.
    :param TiTv: transition/transversion ratio. If ``None``, estimated
        by maximum likelihood. Otherwise, must be a stricly positive
        value. Ignored if data are not nucleotides or if the model does
        not support it. For the ``TN93`` model, there must be a pair of
        ratios, one for purines and one for pyrimidines (in that order).
        However, a single value can be supplied (it will be applied to
        both rates).
    :param pinv: proportion of invariable sites. If ``None`` estimated
        by maximum likelihood. Otherwise, must be in the range [0, 1].
    :param alpha: gamma shape parameter. If ``None``, estimated by
        maximum likelihood. Otherwise, must be a strictly positive
        value. Ignored if *rates* is 1 or if *free_rates* is ``True``.
    :param use_median: boolean indicating whether the median (instead of
        the mean) should be use to report values for rate from the
        discretized gamma distribution. Ignored if *rates* is 1 or if
        *free_rates* is ``True``.
    :param free_rates: boolean indicating whether a mixture model should
        be used for substitution rate categories instead of the
        discretized gamma. In this case all reates and their frequencies
        will be estimated. Requires that *rates* is larger than 1.
    :param seed: pseudo-random number generator seed. Must be a stricly
        positive integer, preferably large.
    :param verbose: boolean indicating whether standard output of PhyML
        should be displayed.

    :return: A ``(tree, stats)`` :class:`tuple` where *tree* is a :class:`.Tree`
        instance and *stats* is a :class:`dict` containing the following
        statistics or estimated parameters:

            * ``lk`` -- log-likelihood of the returned tree and model.
            * ``pars`` -- parsimony score of the returned tree.
            * ``size`` --  length of the returned tree.
            * ``rates`` -- only available if model is ``GTR`` or
              custom, only for nucleotide sequences: relative
              substitution rates, as a :class:`list` providing values in
              the following order:

              #. A :math:`\leftrightarrow` C,
              #. A :math:`\leftrightarrow` G
              #. A :math:`\leftrightarrow` T
              #. C :math:`\leftrightarrow` G
              #. C :math:`\leftrightarrow` T
              #. G :math:`\leftrightarrow` T
            * ``alpha`` -- gamma shape parameter (only if the number of
              rate categories is larger than 1 and if ``free_rates`` was
              ``False``).
            * ``cats`` -- list of ``(rate, proportion)`` tuples for each
              discrete rate category (only if ``free_rates`` was
              ``True``, implying that the number of rates was larger
              than 1).
            * ``freq`` -- list of the relative base frequencies, in the
              following order: A, C, G, and T (only for nucleotide
              sequences).
            * ``ti/tv`` -- transition/transversion ratio (available for
              the models ``K80``, ``HKY85``, ``F84``, and ``TN93``). For
              the ``TN93`` model, the resulting value is a pair of
              transition/transversion ratios, one for purines and one
              for pyrimidines (in that order).
            * ``pinv`` -- proportion of invariable sites (only if the
              corresponding option was not set to 0).

    The choice of the model defines the type of data that are expected.
    The available models are:

    * Nucleotides:

        ========= ============================= ===== ================
        Code      Full name                     Rates Base frequencies
        ========= ============================= ===== ================
        ``JC69``  Jukes and Cantor 1969         one   equal
        ``K80``   Kimura 1980                   two   equal
        ``F81``   Felsenstein 1981              one   unequal
        ``HKY85`` Hasegawa, Kishino & Yano 1985 two   unequal
        ``F84``   Felsenstein 1984              two   unequal
        ``TN93``  Tamura and Nei 1993           three unequal
        ``GTR``   general time reversible       six   unequal
        ========= ============================= ===== ================

        In addition, custom nucleotide substitution models can be
        specified. In that case, *model* must be a six-character
        strings of numeric characters specifying which of the six
        (reversable) substitution rates are allowed to vary. The
        one-rate model is specified by the string ``000000``, the
        two-rate model (separate transition and transversion rates)
        is specified by ``010010``, and the GTR model is specified
        by ``012345``. The substitution rates are specified in the
        following order:

        #. A :math:`\leftrightarrow` C,
        #. A :math:`\leftrightarrow` G
        #. A :math:`\leftrightarrow` T
        #. C :math:`\leftrightarrow` G
        #. C :math:`\leftrightarrow` T
        #. G :math:`\leftrightarrow` T

    * Amino acids:

        ============ ======================================================================
        Code         Authors
        ============ ======================================================================
        ``LG``       Le & Gascuel (*Mol. Biol. Evol.* 2008)
        ``WAG``      Whelan & Goldman (*Mol. Biol. Evol.* 2001)
        ``JTT``      Jones, Taylor & Thornton (*CABIOS* 1992)
        ``MtREV``    Adachi & Hasegawa (*in* *Computer Science Monographs* 1996)
        ``Dayhoff``  Dayhoff *et al.* (*in* *Atlas of Protein Sequence and Structure* 1978)
        ``DCMut``    Kosiol & Goldman (*Mol. Biol. Evol.* 2004)
        ``RtREV``    Dimmic *et al.* (*J. Mol. Evol.* 2002)
        ``CpREV``    Adachi *et al.* (*J. Mol. Evol.* 2000)
        ``VT``       Muller & Vingron (*J. Comput. Biol.* 2000)
        ``Blosum62`` Henikoff & Henikoff (*PNAS* 1992)
        ``MtMam``    Cao *et al.* (*J. Mol. Evol.* 1998)
        ``MtArt``    Abascal, Posada & Zardoya (*Mol. Biol. Evol.* 2007)
        ``HIVw``     Nickle *et al.* (*PLoS One* 2007)
        ``HIVb``     *ibid.*
        ============ ======================================================================

    .. versionchanged:: 3.0.0

        No more default value for *model* option. Added custom model for
        nucleotides. Changed SH pseudo-bootstrap option flag from -3 to
        -4. *quiet* function replaced by *verbose*. Several additional
        options are added. The syntax for input a user tree is modified.
        The second item in the returned tuple is a dictionary of
        statistics.
    """

    # check that program is available
    path = _app.get_path()
    if path is None:
        raise RuntimeError('PhyML program not available -- please configure path')
    command_line = [path, '-i', 'i']

    # check alignment (has enough sequences, ...)
    if not isinstance(align, _interface.Align):
        raise TypeError('`align` should be an Align instance')
    n = align.ns
    if n < 3:
        raise ValueError('alignment does not contain enough sequences to run PhyML')
    if align.ls < 1:
        raise ValueError('alignment does not contain enough data to run PhyML')

    # get model and determines data type
    custom_flag = False
    if model in _aa_models: dt = 'aa'
    elif model in _nt_models: dt = 'nt'
    else:
        if not isinstance(model, str): raise TypeError('`model` must be a string')
        if len(model) != 6: raise ValueError('invalid model name or string')
        try: x = list(map(int, model)) # must translate map to a list because it is used in two different iterations (by set)
        except ValueError: raise ValueError('invalid model name or string')
        if list(range(len(set(x)))) != sorted(set(x)): raise ValueError('invalid custom model specification')
        custom_flag = True
        dt = 'nt'
    command_line.extend(['-d', dt, '-m', model])

    # check that the alignment alphabet is valid for this model
    if dt == 'nt' and align._alphabet._obj.get_type() != 'DNA' and align._alphabet._obj.get_type() != 'codons':
        raise ValueError('provided alignment must contain DNA sequence for model {0}'.format(model))
    if dt == 'aa' and align.alphabet != alphabets.protein:
        raise ValueError('provided alignment must contain protein sequence for model {0}'.format(model))

    # check ti/tv
    if TiTv is None: TiTv = 'e'
    if model == 'TN93':
        if isinstance(TiTv, (float, int)): TiTv = TiTv, TiTv
        else: assert len(TiTv) == 2
        if min(TiTv) <= 0.0: raise ValueError('invalid value for `TiTv` argument')
        command_line.extend(['-t', '{0},{1}'.format(*TiTv)])
    else:
        if TiTv <= 0.0: raise ValueError('invalid value for `TiTv` argument')
        command_line.extend(['-t', '{0}'.format(TiTv)])

    # encode and save alignment
    f = open('i', 'w')
    f.write('{0} {1}\n'.format(n, align.ls*3 if align._alphabet._obj.get_type() == 'codons' else align.ls))
    mapping1 = {} # without labels
    mapping2 = {} # with labels if labels==True (same as mapping1 otherwise)
    for i, item in enumerate(align):
        name = 'in_{0}'.format(i+1)
        if labels:
            mapping1[name] = item.name
            mapping2[name] = '{0}@{1}'.format(item.name, ','.join(map(str, item.labels)))
        else:
            mapping1[name] = item.name
        f.write('{0}  {1}\n'.format(name, item.sequence.string()))
    f.close()
    if not labels: mapping2 = mapping1

    # check boot option
    if not isinstance(boot, int): raise TypeError('`boot` argument must be an int')
    if  boot < -4 and boot != -3: raise ValueError('invalid value for `boot` argument')
    command_line.extend(['-b', str(boot)])

    # manage freq option
    if freq is None:
        if dt == 'nt': freq = 'o'
        else: freq = 'm'
    elif freq == 'o': pass
    elif freq == 'm': pass
    else:
        if len(freq) != 4: raise ValueError('invalid value for `freq` argument')
        if abs(sum(freq) - 1.0) > 0.000000001: raise ValueError('invalid value for `freq` argument')
        freq = ','.join(map(str, freq))
    command_line.extend(['-f', str(freq)])

    # check pinv option
    if pinv is None: pinv = 'e'
    elif pinv < 0.0 or pinv > 1.0: raise ValueError('invalid value for `pinv` argument')
    command_line.extend(['-v', str(pinv)])

    # set gamma distribution parameters
    if not isinstance(rates, int): raise TypeError('`rates` argument must be an int')
    if rates < 1: raise ValueError('invalid value for `rates` argument')
    if alpha is None: alpha = 'e'
    else:
        if not isinstance(alpha, float): raise TypeError('`alpha` argument must be a float')
        if alpha <= 0.0: raise ValueError('invalid value for `alpha` argument')
    command_line.extend(['-c', str(rates), '-a', str(alpha)])
    if free_rates:
        if rates == 1: raise ValueError('cannot use `free_rates` if `rates` < 2')
        command_line.append('--free_rates')

    # input tree
    user_tree_flag = False
    if start_tree == 'nj': pass
    elif start_tree == 'pars': command_line.append('--pars')
    else:
        # if input tree is Tree, make a copy and rename all leaves
        if not isinstance(start_tree, _tree.Tree): raise ValueError('invalid value for `start_tree` argument')

        # try to reverse mapping (will not work if sequences are ambiguous in input)
        rmap = {}
        for k, v in mapping1.items():
            if v in rmap: raise ValueError('cannot use a user tree because of replicates')
            rmap[v] = k

        # rename all leaves
        for leaf in start_tree.iter_leaves():
            lbl = leaf.label
            if lbl not in rmap: raise ValueError('invalid user tree: leaves do not match alignment')
            leaf.label = rmap[lbl]
            del rmap[lbl]
        if len(rmap) > 0: raise ValueError('invalid user tree: not enough leaves')

        # convert tree to string (check that branch lengths are there if required
        if fixed_brlens:
            for node in start_tree.depth_iter():
                for child in node.children():
                    if node.branch_to(child) is None: raise ValueError('`fixed_brlens` requires branch length values')
            f = open('t', 'w')
            f.write(start_tree.newick(skip_labels=True, skip_brlens=False) + '\n')
            f.close()
        else:
            f = open('t', 'w')
            f.write(start_tree.newick(skip_labels=True, skip_brlens=True) + '\n')
            f.close()

        # set argument
        command_line.extend(['-', 't'])
        user_tree_flag = True

    # optimization option
    if fixed_topology or fixed_brlens and not user_tree_flag: raise ValueError('`fixed_topology` option requires a user tree')
    if fixed_brlens: command_line.extend(['-o', 'r'])
    elif fixed_topology: command_line.extend(['-o', 'lr'])
    else: command_line.extend(['-o', 'tlr'])

    # random seed
    if seed is not None:
        if not isinstance(seed, int): raise TypeError('`seed` argument must be an int')
        if seed < 1: raise ValueError('invalid value for `seed` argument')
        command_line.extend(['--r_seed', str(seed)])

    # set other options
    if use_median: command_line.append('--use_median')
    if not verbose: command_line.append('--quiet')
    command_line.append('--no_memory_check')

    # run the program
    p = subprocess.Popen(command_line, stdin=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        stdout=(None if verbose else subprocess.PIPE), universal_newlines=True)

    stdout, stderr = p.communicate('Y\n')

    # check error
    stderr = stderr.strip()
    if len(stderr):
        raise RuntimeError('error while running PhyML: {0}'.format(stderr))

    if os.path.isfile('i_phyml_stats.txt'): statsf = 'i_phyml_stats.txt'
    elif os.path.isfile('i_phyml_stats'): statsf = 'i_phyml_stats'
    else: raise RuntimeError('unknown error while running PhyML (try verbose mode)')

    if os.path.isfile('i_phyml_tree.txt'): treef = 'i_phyml_tree.txt'
    elif os.path.isfile('i_phyml_tree'): treef = 'i_phyml_tree'
    else: raise RuntimeError('unknown error while running PhyML (try verbose mode)')

    # get resulting tree
    try:
        tree = _tree.Tree(fname=treef)
    except ValueError:
        raise RuntimeError('cannot import PhyML output (try verbose mode)')

    # decode tree leaves
    for leaf in tree.iter_leaves():
        leaf.label = mapping2[leaf.label]

    # import statistics
    stats = {}
    f = open(statsf)
    string = f.read()
    f.close()

    # likelihood
    mo = re.search('\. Log-likelihood\:[ \t]+([\-\dEe\.]+)', string)
    if mo:
        try: stats['lk'] = float(mo.group(1))
        except ValueError: raise RuntimeError('cannot parse PhyML output (invalid log-likelihood expression)')
    else:
        raise RuntimeError('no log-likelihood found in PhyML output')

    # parsimony score
    mo = re.search('\. Parsimony\:[ \t]+([\d]+)', string)
    if mo:
        try: stats['pars'] = int(mo.group(1))
        except ValueError: raise RuntimeError('cannot parse PhyML output (invalid parsimony score)')
    else:
        raise RuntimeError('no parsimony score found in PhyML output')

    # tree size
    mo = re.search('\. Tree size\:[ \t]+([\.\dEe\-]+)', string)
    if mo:
        try: stats['size'] = float(mo.group(1))
        except ValueError: raise RuntimeError('cannot parse PhyML output (invalid tree size)')
    else:
        raise RuntimeError('no tree size found in PhyML output')

    # nucleotide frequencies
    if dt == 'nt':
        stats['freqs'] = []
        for base in 'A', 'C', 'G', 'T':
            mo = re.search('f\({0}\)= +([\.\dEe\-]+)'.format(base), string)
            if mo:
                try: stats['freqs'].append(float(mo.group(1)))
                except ValueError: raise RuntimeError('cannot parse PhyML output (invalid nucleotide frequency)')
            else:
                raise RuntimeError('missing frequencies in PhyML output')

    # nucleotide substitution rate
    if model == 'GTR' or custom_flag:
        stats['rates'] = []
        for a, b in [('A', 'C'), ('A', 'G'), ('A', 'T'), ('C', 'G'), ('C', 'T'), ('G', 'T')]:
            mo = re.search('{0} <-> {1} [ \t]+([\.\dEe\-]+)'.format(a, b), string)
            if mo:
                try: stats['rates'].append(float(mo.group(1)))
                except ValueError: raise RuntimeError('cannot parse PhyML output (invalid nucleotide substitution rate)')
            else:
                raise RuntimeError('missing substitution rates in PhyML output')

    # alpha parameter or discrete rates
    if rates > 1:
        if not free_rates:
            mo = re.search('- Gamma shape parameter:[ \t]+([\d\.Ee\-]+)', string)
            if mo:
                try: stats['alpha'] = float(mo.group(1))
                except ValueError: raise RuntimeError('cannot parse PhyML output (invalid gamma shape parameter)')
            else:
                raise RuntimeError('missing gamma shape parameter in PhyML output')
        else:
            stats['cats'] = []
            for i in range(rates):
                mo = re.search('\- Relative rate in class {0}:[ \t]+([\.\dEe\-]+) \[(?:prop|freq)\=([\.\dEe\-]+)\]'.format(i+1), string)
                if mo:
                    r, p = mo.groups()
                    try: stats['cats'].append((float(r), float(p)))
                    except ValueError: raise RuntimeError('cannot parse PhyML output (invalid discrete rates parameter)')
                else:
                    raise RuntimeError('missing discrete rate')

    # ti/tv ratio
    if model in ['K80', 'HKY85', 'F84']:
        mo = re.search('\. Transition/transversion ratio:[ \t]+([\.\dEe\-]+)', string)
        if mo:
            try: stats['ti/tv'] = float(mo.group(1))
            except ValueError: raise RuntimeError('cannot parse PhyML output (invalid Ti/Tv ratio)')
        else:
            raise RuntimeError('missing Ti/Tv ratio in PhyML output')
    if model == 'TN93':
        stats['ti/tv'] = [None, None]
        mo = re.search('\. Transition/transversion ratio for purines:[ \t]+([\.\dEe\-]+)', string)
        if mo:
            try: stats['ti/tv'][0] = float(mo.group(1))
            except ValueError: raise RuntimeError('cannot parse PhyML output (invalid Ti/Tv ratio)')
        else:
            raise RuntimeError('missing Ti/Tv ratio in PhyML output')

        mo = re.search('\. Transition/transversion ratio for pyrimidines:[ \t]+([\.\dEe\-]+)', string)
        if mo:
            try: stats['ti/tv'][1] = float(mo.group(1))
            except ValueError: raise RuntimeError('cannot parse PhyML output (invalid Ti/Tv ratio)')
        else:
            raise RuntimeError('missing Ti/Tv ratio in PhyML output')

    # proportion of fixed sites
    if pinv != 0.0: # including None
        mo = re.search('\. Proportion of invariant:[ \t]+([\.\dEe\-]+)', string)
        if mo:
            try: stats['pinv'] = float(mo.group(1))
            except ValueError: raise RuntimeError('cannot parse PhyML output (invalid fixed sites proportion)')
        else:
            raise RuntimeError('missing fixed sites proportion in PhyML output')

    # return
    return tree, stats



"""
    Copyright 2008-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import re, itertools
from .. import eggwrapper as _eggwrapper
from .. import _interface, alphabets
from . import _reading_frame, _seq_manip

# preload all genetic codes
_codes = {}
for i in range(_eggwrapper.GeneticCode.num_codes()):
    code = _eggwrapper.GeneticCode(i)
    key = code.get_code()
    _codes[key] = code

class Translator(object):
    """
    Class to translate nucleotide sequences to proteins.

    :param code: genetic code identifier (see :ref:`here <genetic-codes>`).
        Required to be an integer among the valid values. The default
        value is the standard genetic code.
    """

    def _help_frame_processor(self, ls, frame):
        # take the frame argument and ls and generate self._frame and self._naa
        if frame is None:
            if ls == 0: self._frame = _reading_frame.ReadingFrame([])
            elif ls % 3 != 0: raise ValueError('alignment length must be a multiple of 3')
            else: self._frame = _reading_frame.ReadingFrame([(0, ls)])
        else:
            if frame.num_needed_bases > ls:
                raise ValueError('reading frame is extending past the end of the sequence')
            self._frame = frame
        self._naa = self._frame.num_codons

    def __init__(self, code=1):
        if code not in _codes: raise ValueError('unknown genetic code: {0}'.format(code))
        self._code = _codes[code]

    def translate_codon(self, codon):
        """
        Translate a single codon. Translation is based on the 
        genetic code defined at construction time. Alternative start 
        codons are not supported.

        :param codon: codon as a three-character string
        :return: The one-letter amino acid code if the codon can be
            translated, or ``-`` if the codon is ``---``, or ``X`` in all
            other cases
            (including all cases with invalid nucleotides).
        """
        return alphabets.protein.get_value(
                    self._code.translate(
                        alphabets.codons.get_code(codon)))

    def translate_sequence(self, sequence, allow_alt=False):
        """
        Translate a sequence.

        :param sequence: a :class:`str`, :class:`~.SequenceView` or
            compatible instance containing codon sequences. Only
            upper-case strings.

        :param frame: a :class:`~.ReadingFrame` instance providing the
            exon positions in the correct frame. By default, a
            non-segmented frame covering all sequences is assumed (in
            case the provided alignment is the coding region and length
            must be a multiple of 3).

        :param allow_alt: a boolean telling whether alternative start
            (initiation) codons should be considered. If ``False``,
            codons are translated as a methionine (M) if, and only if,
            there are among the alternative start codons for the considered
            genetic code and they appear at the first position for the
            sequence.

        :return: A new :class:`str` instance containing translated
            (protein) sequences.
        """
        if isinstance(sequence, str):
            if len(sequence) % 3 != 0: raise ValueError('number of bases must be a multiple of 3')
            sequence = re.findall('...', sequence.upper())
        if allow_alt:
            for i, codon in enumerate(sequence):
                if codon == '---': continue
                if self._code.start(alphabets.codons.get_code(codon)):
                    start = i
                    break
                if alphabets.codons.get_code(codon) < 0:
                    start = None
                    break
            else: start = None
        dest = [self._code.translate(alphabets.codons.get_code(codon)) for codon in sequence]
        if allow_alt and start is not None and dest[start] != 10: dest[start] = 10
        return ''.join(map(alphabets.protein.get_value, dest))

    def translate_align(self, src, allow_alt=False, in_place=False):
        """
        Translate an :class:`.Align` instance.

        :param src: an :class:`!Align` containing codon sequences and
            using the :py:obj:`.alphabets.codons` alphabet.

        :param allow_alt: a boolean telling whether alternative start
            (initiation) codons should be considered. If ``True``,
            codons are translated as a methionine (M) if
            they are among the alternative start codons for the considered
            genetic code and they appear at the first position for the
            considered sequence (excluding all triplets of gap symbols
            appearing at the 5' end of the sequence).

        :param in_place: place translated sequences into the original
            :class:`!Align` instance (this discards original data and
            replaces the original alphabet by the protein alphabet). By
            default, returns a new instance.

        :return: By default, an original :class:`!Align` instance
            containing translated (protein) sequences. If *in_place* was
            ``True``, return ``None``.
        """

        # check input
        if not isinstance(src, _interface.Align): raise TypeError('expect an Align instance')
        if src._alphabet._obj.get_type() != 'codons': raise ValueError('alphabet must be codons')
        ns = src.ns
        ls = src.ls

        # create destination
        if not in_place:
            dest = _interface.Align(nsam=ns, nsit=ls, alphabet=alphabets.protein)
            for i in range(ns):
                dest.set_name(i, src.get_name(i))
                dest._obj.set_nlabels(i, src._obj.get_nlabels(i))
                for j in range(src._obj.get_nlabels(i)):
                    dest.set_label(i, j, src.get_label(i, j))
        else:
            dest = src
            dest._alphabet = alphabets.protein

        # iterate over sequences
        for i in range(ns):
            if allow_alt:
                for j in range(ls):
                    codon = src._obj.get_sample(i, j)
                    if codon == -1165: continue
                    if self._code.start(codon):
                        start = j
                        break
                    if codon < 0:
                        start = None
                        break
                else: start = None
            for j in range(ls):
                dest._obj.set_sample(i, j, self._code.translate(src._obj.get_sample(i, j)))
            if allow_alt and start is not None and dest._obj.get_sample(i, start) != 10: dest._obj.set_sample(i, start, 10)

        # return
        if not in_place: return dest

    def translate_container(self, src, allow_alt=False, in_place=False):
        """
        Translate a :class:`.Container` instance.

        :param align: a :class:`!Container` containing codons sequences.

        :param allow_alt: a boolean telling whether alternative start
            (initiation) codons should be considered. If ``False``,
            codons are translated as a methionine (M) if, and only if,
            they are among the alternative start codons for the considered
            genetic code and they appear at the first position for the
            considered sequence.

        :param in_place: place translated sequences into the original
            :class:`!Container` instance (this discards original data).
            By default, return a new instance.

        :return: By default, a new :class:`!Container` instance
            containing translated (protein) sequences. If *in_place* was
            ``True``, return ``None``.
        """

        # check input
        if not isinstance(src, _interface.Container): raise TypeError('expect a Container instance')
        if src._alphabet._obj.get_type() != 'codons': raise ValueError('alphabet must be codons')
        ns = src.ns

        # create destination
        if not in_place:
            dest = _interface.Container(alphabet=alphabets.protein)
            dest._obj.set_nsam(ns)
            dest._ns = ns
            for i in range(ns):
                dest.set_name(i, src.get_name(i))
                dest._obj.set_nlabels(i, src._obj.get_nlabels(i))
                for j in range(src._obj.get_nlabels(i)):
                    dest.set_label(i, j, src.get_label(i, j))
        else:
            dest = src
            dest._alphabet = alphabets.protein

        # iterate over sequences
        for i in range(ns):
            ls = src._obj.get_nsit_sample(i)
            if not in_place: dest._obj.set_nsit_sample(i, ls)
            if allow_alt:
                for j in range(ls):
                    codon = src._obj.get_sample(i, j)
                    if codon == -1165: continue
                    if self._code.start(codon):
                        start = j
                        break
                    if codon < 0:
                        start = None
                        break
                    else: start = None
            for j in range(ls):
                dest._obj.set_sample(i, j, self._code.translate(src._obj.get_sample(i, j)))
            if allow_alt and start is not None and dest._obj.get_sample(i, start) != 10: dest._obj.set_sample(i, start, 10)

        # return
        if not in_place: return dest

def translate(seq, code=1, allow_alt=False, in_place=False):
    """
    Translates coding nucleotide sequences to proteins. See the class
    :class:`.tools.Translator` for more details. This is a convenience method
    allowing to translate nucleotide sequences in a single call. For
    repetitive calls, direct use of :class:`!Translator` can be more
    efficient.

    :param seq: input codon sequences. Accepted types are
        :class:`.Align`, :class:`.Container`, :class:`.SequenceView`
        or :class:`str` (or also types compatible with :class:`str`).
        If the input object is an :class:`!Align` or a :class:`!Container`,
        the alphabet must be :py:obj:`.alphabets.codons`.

    :param code: genetic code identifier (see below).
        Required to be an integer among the valid values. The default
        value is the standard genetic code.

    :param allow_alt: a boolean telling whether alternative start
        (initiation) codons should be considered. If ``True``,
        codons are translated as a methionine (M) if, and only if,
        there are among the alternative start codons for the considered
        genetic code and they appear at the first position for the
        sequence. If *seq* is an :class:`!Align`, leading gaps are
        ignored as long as they appear as multiples of 3 (fully missing codons).

    :param in_place: place translated sequences in the provided
        :class:`!Align` or :class:`!Container` instance, overwriting
        initial data. Not allowed if *seq* is not of one of these two
        types. Replace the original alphabet by :py:obj:`.alphabets.protein`.
        By default, return a new instance.

    :return: Protein sequences as an :class:`!Align` or a
        :class:`!Container` if either of these types have been provided
        as *seq*, or as a string otherwise. If *in_place* has been
        specified, return ``None``.

    .. _genetic-codes:

    All genetic codes defined by the National Center for Biotechnology
    Information are supported and can be accessed using codes corresponding
    to the GenBank ``/trans_table`` qualifier. The codes are integers.

    .. include:: list_genetic_codes.txt

    Reference: National Center for Biotechnology Information
    [`<http://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi>`_]
   """

    T = Translator(code=code)
    if isinstance(seq, _interface.Align):
        return T.translate_align(seq, allow_alt=allow_alt, in_place=in_place)
    elif isinstance(seq, _interface.Container):
        return T.translate_container(seq, allow_alt=allow_alt, in_place=in_place)
    else:
        if in_place == True: raise ValueError('cannot translate sequence in place')
        return T.translate_sequence(seq, allow_alt=allow_alt)

class orf_iter:
    """
    Iterate over open reading frames.
    Return an iterator over non-segmented open reading frames (ORFs)
    found in a provided codon sequence in any of the six possible frames.

    :param sequence: a string or a :class:`.SequenceView` representing a codon sequence.

    :param code: genetic code identifier (see :ref:`here <genetic-codes>`).
        Required to be an integer among the valid values. The default
        value is the standard genetic code.

    :param min_length: minimum length of returned ORFs. This value must
        be at least 1. It is understood as the length of the encoded
        peptide (since ORFs are returned as nucleotide sequences, they
        will have a length of least three times this value).

    :param forward_only: consider only the three forward frames (do not
        consider the reverse strand).

    :param force_start: if ``True``, all returned ORFs are required to
        start with a start codon. Otherwise, all non-stop codons are
        included in ORFs. Note that alternative start codons (``CTG``
        and ``TTG`` for the standard genetic code) are also supported.

    :param allow_alt: allow alternative start codons (only considered
        if *force_start* is set).

    :param force_stop: require that ORFs
        end with a stop codon (this only excludes 3'-partial ORFs, that
        is ORFs that end with the end of the provided sequennce).

    :return: An iterator over all detected ORFs. Each ORF is represented
        by a ``(start, stop, length, frame)`` tuple where ``start`` is the
        start position of the ORF and ``stop`` the stop position (such as
        ``sequence[start:stop]`` returns the ORF sequence or its reverse
        complement [note that the stop codon, if present, is included]),
        ``length`` is the ORF length (number of amino acids, excluding any
        stop codon) and ``frame`` is the reading frame on which it was found: +1, +2,
        +3 are the frames on the forward strand (starting respectively at
        the first, second, and third base), and -1, -2, -3 are the frames
        on the reverse strand (starting respectively at the last, last
        but one, and last but two bases).
    """

    def __init__(self, sequence, code=1, min_length=1,
                 forward_only=False, force_start=True, allow_alt=False,
                 force_stop=True):

        # get/check arguments
        if code not in _codes: raise ValueError('unknown genetic code: {0}'.format(code))
        code = _codes[code]
        if min_length < 1: raise ValueError('value for `min_length` is too small')
        self._ls = len(sequence)
        self._sequence = sequence
        self._code = code
        self._min_length = min_length
        self._forward_only = forward_only
        self._force_start = force_start
        self._force_stop = force_stop
        self._frames = [None] * 6

        # make a list of codon status (1:start/2:stop/3:valid/4:missing)
        self._codon_table = {}
        aa_stop = alphabets.protein.get_code('*')
        valid, missing = alphabets.codons.get_alleles()
        for cod in valid:
            c = alphabets.codons.get_code(cod)
            if cod == 'ATG' or (allow_alt and code.start(c)):
                self._codon_table[cod] = 1
            elif code.translate(c) == aa_stop:
                self._codon_table[cod] = 2
            else:
                self._codon_table[cod] = 3
        for cod in missing:
            self._codon_table[cod] = 4

    def __iter__(self):
        self._pos = 0
        self._iter_f = itertools.cycle([1, 2, 3])
        self._iter_r = itertools.cycle([-1, -3, -2])
        if self._ls % 3 == 1:
            next(self._iter_r)
            next(self._iter_r)
        elif self._ls % 3 == 2:
            next(self._iter_r)
        self._cache = None
        return self

    def __next__(self):

        # return cache frame if any
        if self._cache:
            r = self._cache
            self._cache = None
            return r

        # proceed until an ORF (or the end) is found
        while True:

            # if reached end of sequence, process hanging ORFs and finish
            if self._pos > self._ls-3:
                return self.finishing() # ultimately raises StopIteration

            orf_f = None
            orf_r = None

            # process forward frame
            F = next(self._iter_f)
            cod = self._codon_table[self._sequence[self._pos:self._pos + 3]]
            if self._frames[F-1] is None:
                if (self._force_start == True and cod == 1) or (self._force_start == False and cod in [1, 3]):
                    self._frames[F-1] = [self._pos, None, None, F]
            else:
                if cod == 2:
                    self._frames[F-1][1] = self._pos + 3
                    self._frames[F-1][2] = (self._pos + 3 - self._frames[F-1][0]) // 3 - 1 # -1 because of stop
                    if self._frames[F-1][2] >= self._min_length: orf_f = tuple(self._frames[F-1])
                    self._frames[F-1] = None

            # process reverse frame
            if not self._forward_only:
                R = next(self._iter_r)
                cod = self._codon_table[_seq_manip.rc(self._sequence[self._pos:self._pos + 3])]
                if self._frames[2-R] is None:
                    if cod == 2:            # start, stop, len, frame, final_stop, last_M, has_non_stop
                        self._frames[2-R] = [self._pos, None, None, R, 1, None, False]
                    elif cod != 4 and not self._force_stop:
                        self._frames[2-R] = [self._pos, None, None, R, 0, self._pos + 3 if cod==1 else None, True]
                else:
                    if self._force_start == True and cod == 1:
                        self._frames[2-R][5] = self._pos + 3
                        self._frames[2-R][6] |= True
                    elif cod == 2:
                        if self._force_start:
                            if self._frames[2-R][5] is not None: # there is an M
                                self._frames[2-R][1] = self._frames[2-R][5]
                                self._frames[2-R][2] = (self._frames[2-R][5] - self._frames[2-R][0]) // 3 - self._frames[2-R][4]
                                if self._frames[2-R][2] >= self._min_length: orf_r = tuple(self._frames[2-R][:4])
                        elif self._frames[2-R][6]:
                            self._frames[2-R][1] = self._pos
                            self._frames[2-R][2] = (self._pos - self._frames[2-R][0]) // 3  - self._frames[2-R][4]
                            if self._frames[2-R][2] >= self._min_length: orf_r = tuple(self._frames[2-R][:4])
                        self._frames[2-R] = [self._pos, None, None, R, 1, None, False]
                    elif cod != 4:
                        self._frames[2-R][6] |= True

            # return frame (return orf_f and cache orf_r if both)
            self._pos += 1
            if orf_f and orf_r:
                self._cache = orf_r
                return orf_f
            if orf_f:
                return orf_f
            if orf_r:
                return orf_r

    def finishing(self):
        # if we reached the end, return any remaining still-running ORF (if allowed by options)

        # forward frames
        if not self._force_stop:
            for i in range(3):
                if self._frames[i] is not None:
                    self._frames[i][1] = (self._ls - i) // 3 * 3 + i
                    self._frames[i][2] = (self._frames[i][1] - self._frames[i][0]) // 3
                    r = self._frames[i]
                    self._frames[i] = None
                    if r[2] >= self._min_length: return tuple(r)

        # reverse frames
        for i in range(3):
            if self._frames[3+i] is not None:
                if self._force_start:
                    if self._frames[3+i][5] is None:
                        self._frames[3+i] = None
                        continue
                    self._frames[3+i][1] = self._frames[3+i][5]
                    self._frames[3+i][2] = (self._frames[3+i][5] - self._frames[3+i][0]) // 3 - self._frames[3+i][4]
                elif self._frames[3+i][6]:
                    self._frames[3+i][1] = self._ls - i
                    self._frames[3+i][2] = (self._frames[3+i][1] - self._frames[3+i][0]) // 3 - self._frames[3+i][4]
                if self._frames[3+i][6]:
                    r = self._frames[3+i][:4]
                    self._frames[3+i] = None
                    if r[2] >= self._min_length: return tuple(r)
                else:
                    self._frames[3+i] = None

        raise StopIteration

def longest_orf(*args, **kwargs):
    """
    longest_orf(sequence, code=1, min_length=1,
                 forward_only=False, force_start=True, allow_alt=False,
                 force_stop=True)

    Detect the longest open reading frame. Arguments are identical to :func:`.orf_iter`.
    A :exc:`ValueError` is raised if two or more open
    reading frames have the largest length.

    :return: A ``(start, stop, length, frame)`` tuple (see the return
        value of :func:`.orf_iter` for details), or ``None`` if no open
        reading frame fits the requirements.
    """
    max_length = 0
    candidate = None
    for start, stop, length, frame in orf_iter(*args, **kwargs):
        if length == max_length:
            candidate = None
            num += 1
        elif length > max_length:
            max_length = length
            candidate = start, stop, length, frame
            num = 1

    if max_length == 0: return None
    if num > 1: raise ValueError('several equally long ORFs found')
    return candidate

def backalign(nucl, aln, code=1, ignore_names=False, ignore_mismatches=False, fix_stop=False, allow_alt=True):
    """
    Align coding sequence based on the protein alignment.

    :param nucl: a :class:`.Container` or :class:`.Align` instance
        containing coding sequences that should be aligned. There should
        not be any alignment gap in the sequences. The alphabet must be
        :class:`.alphabets.codons`.

    :param aln: an :class:`!Align` instance containing an alignment of
        the protein sequences encoded by the coding sequences provided
        as *nucl*. Group labels of *aln*
        are not taken into account. The alphabet must be :class:`.alphabets.protein`.

    :param code: genetic code identifier (see :ref:`here <genetic-codes>`).
        Required to be an integer among the valid values. The default
        value is the standard genetic code.

    :param ignore_names: if ``True``, ignore names for matching
        sequences in the protein alignment to coding sequences.
        Sequences will be matched using their rank and the names in the
        returned alignment will be taken from *nucl*.

    :param ignore_mismatches: if ``True``, do not generate any exception
        if a predicted protein does not match the provided protein
        sequence (if the lengths differ, an exception is always raised).

    :param fix_stop: if ``True``, support a single trailing stop codon
        in coding sequences not represented by a ``*`` in the provided
        protein alignment (if the final stop codons have been
        stripped during alignment). If found, this stop codon will be
        flushed as left as possible (immediately after the last non-gap
        character) in the returned coding alignment.

    :return: An :class:`.Align` instance containing aligned coding codon
        sequences.

    If a mismatch is detected between a protein from *aln* and the
    corresponding prediction from *nucl*, an instance of
    :exc:`.tools.BackalignError` (a subclass of :exc:`ValueError`)
    is raised. Its attribute :py:obj:`~.tools.BackalignError.alignment` can be
    used to help identify the reason of the error. Mismatches (but not
    differences of length) can be ignored with the option
    *ignore_mismatches*.
    """
    # checking
    if not isinstance(aln, _interface.Align): raise TypeError('expect an Align instance')
    if aln._alphabet._obj.get_name() != 'protein': raise ValueError('alignment alphabet must be protein')
    if not isinstance(nucl, _interface.Container): raise TypeError('argument must be a Container instance')
    if nucl._alphabet._obj.get_type() != 'codons': raise ValueError('alignment alphabet must be codons')

    # get code
    translator = Translator(code=code)

    # initialize the return object
    ns = aln.ns
    ls = aln.ls
    result = _eggwrapper.DataHolder(True)
    result.set_nsam(ns)
    result.set_nsit_all(ls + int(fix_stop))

    # set a trailing gape at the end of all sequences to allow 1 or more additional stop codons
    if fix_stop:
        for i in range(ns): result.set_sample(i, ls, -1165)
        last_used = False # set to True if at least one sequence uses the last three bases

    # get mapping of indexes
    nucl = nucl._obj
    aln = aln._obj
    if nucl.get_nsam() != ns: raise ValueError('numbers of sequences don\'t match')
    names = list(map(nucl.get_name, range(ns)))
    if not ignore_names:
        if len(set(names)) < ns: raise ValueError('duplicate name found')
        aln_names = list(map(aln.get_name, range(ns)))
        ranks = []
        for name in names:
            try: ranks.append(aln_names.index(name))
            except ValueError: 'sequence name not found in protein alignment: {0}'.format(name)
    else:
        ranks = list(range(ns))

    # set names and group labels
    for idx_cds in range(ns):
        result.set_name(idx_cds, names[idx_cds])
        result.set_nlabels(idx_cds, nucl.get_nlabels(idx_cds))
        for k in range(nucl.get_nlabels(idx_cds)):
            result.set_label(idx_cds, k, nucl.get_label(idx_cds, k))

    # main loop
    last_used = False
    for idx_cds, idx_aln in enumerate(ranks):
        ls_nucl = nucl.get_nsit_sample(idx_cds)
        if ls_nucl < 1: raise ValueError('empty nucleotide sequence')

        # process all aligned aa positions
        c = 0
        for idx in range(ls):
            aa = aln.get_sample(idx_aln, idx)

            # process gaps
            if aa == -1: result.set_sample(idx_cds, idx, -1165)

            # non-gap
            else:
                codon = nucl.get_sample(idx_cds, c)
                c += 1
                if (not ignore_mismatches and translator._code.translate(codon) != aa): # check mismatch
                    raise BackalignError(names[idx_cds], nucl.get_sample, aln.get_sample, idx_cds, idx_aln, ls, ls_nucl, translator)
                result.set_sample(idx_cds, idx, codon) # set codon

        # if there is an additional codon at the end
        if c != ls_nucl:

            # if requested, check that it is a stop codon
            if fix_stop:
                codon = nucl.get_sample(idx_cds, c)
                c += 1

                # fix stop codon (only if all bases have been read)
                if c == ls_nucl and translator._code.translate(codon) == 20:
                    i = ls # there is necessarily room there, filled by gaps
                    while i > 0 and result.get_sample(idx_cds, i-1) == -1165: i -= 1
                    result.set_sample(idx_cds, i, codon)
                    if i == ls: last_used = True
                else: raise BackalignError(names[idx_cds], nucl.get_sample, aln.get_sample, idx_cds, idx_aln, ls, ls_nucl, translator)
            else: raise BackalignError(names[idx_cds], nucl.get_sample, aln.get_sample, idx_cds, idx_aln, ls, ls_nucl, translator)

    # if the last codon position has not been used at least once by fix_stop, remove them
    if fix_stop and not last_used:
        result.set_nsit(ls) # decrease by three

    # that's all, folks!
    return _interface.Align._create_from_data_holder(result, alphabets.codons)

class BackalignError(ValueError):
    """
    Exception used to report errors
    occurring during the use of :func:`.tools.backalign` because of mismatches
    between the provided alignment and predicted proteins.
    """

    def __init__(self, name, fnuc, faln, i_nuc, i_aln, ls_aln, ls_nuc, translator):
        self._name = name
        message = 'mismatch between provided and predicted proteins for `{0}`'.format(name)
        prov = ''.join(map(alphabets.protein.get_value, [faln(i_aln, i) for i in range(ls_aln)])).replace('-', '')
        pred = ''.join(map(alphabets.codons.get_value, [fnuc(i_nuc, i) for i in range(ls_nuc)])).replace('-', '')
        if len(pred) % 3 != 0: message = 'length of nucleotide sequence not a multiple of 3 for {0}'.format(name)
        pred = ''.join([translator.translate_codon(''.join(codon)) for codon in zip(pred[::3], pred[1::3], pred[2::3])])

        n = max([len(prov), len(pred)])

        mid = []
        for i in range(n):
            if i >= len(prov):
                prov += '-'
                mid.append('~')
            elif i >= len(pred):
                pred += '-'
                mid.append('~')
            elif prov[i] != pred[i]:
                mid.append('#')
            else:
                mid.append('|')
        mid = ''.join(mid)

        c = 0
        self._alignment = []
        while True:
            A = prov[c:c+60]
            A = ' '.join([A[i*10:i*10+10] for i in range(6)])
            B = mid[c:c+60]
            B = ' '.join([B[i*10:i*10+10] for i in range(6)])
            C = pred[c:c+60]
            C = ' '.join([C[i*10:i*10+10] for i in range(6)])
            self._alignment.append('[provided]  {0}'.format(A) + '[{0}]\n'.format(c+60).rjust(7))
            self._alignment.append('            {0}\n'.format(B))
            self._alignment.append('[predicted] {0}'.format(C) + '[{0}]\n'.format(c+60).rjust(7))
            c += 40
            if c < n: self._alignment.append('\n')
            else: break

        self._alignment = ''.join(self._alignment)
        ValueError.__init__(self, message)

    @property
    def name(self):
        """
        Name of sequence for which the error occurred.
        """
        return self._name

    @property
    def alignment(self):
        """
        Comparaison of provided and predicted proteins.
        This string shows the two proteins with a middle line composed
        of the following symbols:

        * ``|``: match.
        * ``#``: mismatch.
        * ``~``: one protein shorter.
        """
        return self._alignment

def trailing_stops(align, action=0, code=1, replacement='???'):
    """
    Process stop codons at the end of the sequences. This function
    detects and (optionally) fix trailing stop codons.
    The algorithm consists in locating the last non-gap codon of each
    sequence. If the last codon is partially gapped (which can happen
    if coding sequences are aligned using a DNA-based algorithm), it
    will not be detected and ignored.

    :param align: an :class:`.Align` containing aligned coding
        sequences. The alphabet must be :py:obj:`.alphabets.codons`.

    :param action: an integer specifying what should be done if a stop
        codon is found at the end of a given sequence. Possible actions
        are listed in the following table:

        +------+---------------------------------------------------+
        | Code | Action                                            +
        +======+===================================================+
        | 0    | Nothing (just count them).                        +
        +------+---------------------------------------------------+
        | 1    | Replace them by a gap, and delete the last        +
        |      | position if it is made by gaps only.              +
        +------+---------------------------------------------------+
        | 2    | Replace them by the value given as *replacement*. +
        +------+---------------------------------------------------+

        Note that using ``action=1`` is not stricly equivalent to using
        ``action=2, replacement='---'`` because the former deletes the
        last three positions of the alignment if needed while the latter
        does not.

    :param code: genetic code identifier (see :ref:`here <genetic-codes>`).
        Required to be an integer among the valid values. The default
        value is the standard genetic code.

    :param replacement: if *action* is set to 2, provide the codon
        that should be used to replace stop codons.

    :return: The number of sequences that had a trailing stop codons
        among the considered sequences.
    """

    # argument check
    if not isinstance(align, _interface.Align): raise TypeError('expect an Align instance')
    if align._alphabet.type != 'codons': raise ValueError('alignment alphabet must be codons')
    if code not in _codes: raise ValueError('unknown genetic code: {0}'.format(code))
    code = _codes[code]
    ns = align.ns
    ls = align.ls
    try: replacement = alphabets.codons.get_code(replacement)
    except ValueError: raise ValueError('invalid codon: {0}'.format(replacement))
    cnt = 0
    num_gapped = 0

    for i in range(ns):
        cur = ls - 1
        for cur in range(ls-1, -1, -1):
            if align._obj.get_sample(i, cur) != -1165:
                if code.translate(align._obj.get_sample(i, cur)) == 20:
                    cnt += 1
                    if action == 1:
                        align._obj.set_sample(i, cur, -1165)
                        num_gapped += 1
                    elif action == 2:
                        align._obj.set_sample(i, cur, replacement)
                break

    # remove final 3 gaps (only if ALL samples where replaced by gaps, in particular only if include_outgroup was true)
    if action == 1 and num_gapped == ns:
        align._obj.set_nsit_all(ls-3)

    return cnt

def iter_stops(src, code=1):
    """
    Iterate over stop codons.
    Return an iterator providing the coordinates of all stop codons
    over all sequences. Each iteration returns a
    ``(sample, position)`` where ``sample`` is the sample index and
    ``position`` is the index of the stop codon.

    :param src: an :class:`.Align` or :class:`.Container` containing
        coding sequences. The alphabet must be :py:obj:`.alphabets.codons`.

    :param code: genetic code identifier (see :ref:`here <genetic-codes>`).
        Required to be an integer among the valid values. The default
        value is the standard genetic code.

    :return: An iterator over the ``(sample, position)`` tuples
        corresponding to each stop codon found in the alignment (see
        above).
    """

    # argument check
    if not isinstance(src, (_interface.Align, _interface.Container)): raise TypeError('expect an Align or Container instance')
    if src._alphabet.type != 'codons': raise ValueError('alignment alphabet must be codons')
    ns = src.ns
    if code not in _codes: raise ValueError('unknown genetic code: {0}'.format(code))
    code = _codes[code]
    if isinstance(src, _interface.Align): ls = src.ls

    for i in range(ns):
        if isinstance(src, _interface.Container): ls = src.ls(i)
        for j in range(ls):
            if code.translate(src._obj.get_sample(i, j)) == 20:
                yield (i, j)

def has_stop(src, code=1):
    """
    Test if there is at least one stop codon.
    Return ``True`` if the sequences contain at least one codon stop
    at any position in any sequence, and ``False`` otherwise.

    :param src: an :class:`.Align` or :class:`.Container` containing
        aligned coding sequences.

    :param code: genetic code identifier (see :ref:`here <genetic-codes>`).
        Required to be an integer among the valid values. The default
        value is the standard genetic code.

    :return: A boolean.
    """
    for stop in iter_stops(src, code=code): break
    else: return False
    return True



"""
    Copyright 2008-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from .. import alphabets

class ReadingFrame(object):
    """
    Handle reading frame positions. The reading frame positions can be
    loaded as constructor argument or using the method
    :meth:`~.ReadingFrame.process`. By default, build an instance with
    no exons.

    .. _reading-frame-constructor-arguments:

    :param frame: the reading frame specification must be a sequence
        of ``(start, stop[, codon_start])`` pairs or triplets where
        *start* and *stop* give the limits of an exon, such that
        ``sequence[start:stop]`` returns the exon sequence, and
        *codon_start*, if specified, can be:
        
        * 1 if the first position of the exon is the first position
          of a codon (e.g. ``ATG ATG``),

        * 2 if the first position of the segment is the second
          position of a codon (e.g. ``TG ATG``),

        * 3 if the first position of the segment is the third
          position a of codon (e.g. ``G ATG``),

        * ``None`` if the reading frame is continuing the previous
          exon.

    :param keep_truncated: if ``True``, codons that are truncated
        (either because the reading frame is 5', internally, or 3'
        partial, or if the number of bases is not a multiple of
        three) are not skipped (missing positions of truncated
        codons are replaced by ``None``).

    If *codon_start* of the first segment is ``None``, 1 will be
    assumed. If *codon_start* of any non-first segment is not
    ``None``, the reading frame is supposed to be interupted. This
    means that if any codon was not completed at the end of the
    previous exon, it will assumed to be assumed incomplete, and the first
    codon of the current codon will be incomplete as well.
    """

    def __init__(self, frame=None, keep_truncated=False):
        if frame is None: frame = []
        self.process(frame, keep_truncated)

    def process(self, frame, keep_truncated=False):
        """
        Reset instance with a new reading frame. All previously loaded data are discarded.
        The arguments of this method are identical to the :ref:`constructor <reading-frame-constructor-arguments>`.
        """
        cur = -1
        self._starts = []
        self._stops = []
        self._codon_starts = []
        for exon in frame:
            if len(exon) == 2:
                start, stop = exon
                codon_start = None
            elif len(exon) == 3:
                start, stop, codon_start  = exon
                if codon_start not in set([1, 2, 3, None]): raise ValueError('invalid value for `codon_start`: {0}'.format(codon_start))
            if start < cur: raise ValueError('exon limits must in increasing order')
            if stop <= start: raise ValueError('exon limits must in increasing order')
            cur = stop
            self._starts.append(start)
            self._stops.append(stop)
            self._codon_starts.append(codon_start)

        self._num_exons = len(self._starts)
        if self._num_exons > 0:
            self._num_tot_bases = self._stops[-1] - self._starts[0]
            self._num_exon_bases = sum([stop-start for (start, stop) in zip(self._starts, self._stops)])
            if self._codon_starts[0] is None: self._codon_starts[0] = 1 # important
        else:
            self._num_tot_bases = 0
            self._num_exon_bases = 0

        self._codons = []
        self._bases = {} # keys are bases, values are [exon, codon] indexes pairs
        for idx, (start, stop, codon_start) in enumerate(zip(self._starts, self._stops, self._codon_starts)):
            if codon_start != None:
                # if codon_start is specified, complete the previous codon
                if len(self._codons):
                    if keep_truncated:
                        self._codons[-1] += [None] * (3 - len(self._codons[-1]))
                    else:
                        if len(self._codons[-1]) != 3: del self._codons[-1]

                # add non codon with shift if needed
                if keep_truncated:
                    if codon_start == 1: self._codons.append([])
                    elif codon_start == 2: self._codons.append([None])
                    elif codon_start == 3: self._codons.append([None, None])
                    else: raise ValueError('invalid value for codon start')
                else:
                    self._codons.append([])
                    if codon_start == 1: pass
                    elif codon_start == 2:
                        self._bases[start] = [idx, None] # avoid side effect of deleting bases due to shift
                        self._bases[start+1] = [idx, None]
                        start += 2
                    elif codon_start == 3:
                        self._bases[start] = [idx, None]
                        start += 1
                    else: raise ValueError('invalid value for codon start')

            # loop over bases, adding codons when the current is filled
            # it is guaranteed that there will always be one codon (because 1st exon is in frame 1)
            for i in range(start, stop):
                if len(self._codons[-1]) == 3: self._codons.append([])
                self._codons[-1].append(i)
                self._bases[i] = [idx, None]

        # complete last codon if needed
        if len(self._codons) > 0:
            if keep_truncated:
                self._codons[-1] += [None] * (3 - len(self._codons[-1]))
            else:
                if len(self._codons[-1]) != 3: del self._codons[-1]

        # final processing
        self._codons = tuple(map(tuple, self._codons))
        for idx, codon in enumerate(self._codons):
            for i in codon:
                if i is not None:
                    if i not in self._bases or self._bases[i][1] is not None:
                        raise RuntimeError('error in ReadingFrame code')
                    self._bases[i][1] = idx
        if self._num_exons > 0:
            self._needed_bases = self._stops[-1]
        else: self._needed_bases = 0

    @property
    def num_needed_bases(self):
        """
        Number of bases needed to apply this reading
        frame. Minimum size of a sequence to be consistent with this
        reading frame. In practice, the value equals to the end of the last exon
        or zero if there are no exons. If the reading frame is used with
        a shorter sequence, it can lead to errors.
        """
        return self._needed_bases

    @property
    def num_tot_bases(self):
        """
        Number of bases of the reading frame. All introns and exons are
        included, starting from the start of the first exon
        up to end of the last one.
        """
        return self._num_tot_bases

    @property
    def num_exon_bases(self):
        """ Number of bases in exons. """
        return self._num_exon_bases

    @property
    def num_exons(self):
        """ Number of exons. """
        return self._num_exons

    @property
    def num_codons(self):
        """ Number of codons. Truncated codons are included. """
        return len(self._codons)

    def exon_index(self, base):
        """
        Find the exon in which a given base falls.

        :param base: any base index.
        :return: The index of the corresponding exon, or ``None`` if the
            base does not fall in any exon.
        """
        if base in self._bases: return self._bases[base][0]
        else: return None

    def codon_index(self, base):
        """
        Find the codon in which a given base falls.

        :param base: any base index.
        :return: The index of the corresponding codon, or ``None`` if
            the base does not fall in any codon.
        """
        if base in self._bases: return self._bases[base][1]
        else: return None

    def codon_position(self, base):
        """
        Tell the position of a base within the
        codon in which it falls.

        :param base: any base index.
        :return: The index of the base in the codon (0, 1 or 3), or
            ``None`` if the base does not fall in any codon.
        """
        if base in self._bases and self._bases[base][1] is not None:
            return self._codons[self._bases[base][1]].index(base)
        else: return None

    def codon_bases(self, codon):
        """
        Give the position of the three bases of a given codon. One or
        two positions (but never the middle one alone) will be ``None``
        if the codon is truncated (beginning/end of an exon without
        coverage of the previous/next one). If the codon index is out of
        range, return ``None``.
        
        :param codon: any codon index.
        :return: A tuple with the three base positions (containing potentially
            one or two ``None``) or
            ``None`` if the codon index is out of range.
        """
        try: return self._codons[codon]
        except IndexError: return None

    def iter_exon_bounds(self):
        """
        Iterate over exons.
        This iterator returns ``(start, stop)`` tuples of the positions
        of the limits of each exon.
        """
        for start, stop in zip(self._starts, self._stops): yield start, stop

    def iter_codons(self):
        """
        Iterate over codons.
        This iterator returns ``(first, second, third)`` tuples of the
        positions of the three bases of each codon. Positions might be
        ``None`` if *keep_truncated* has been set to ``True``.
        """
        for codon in self._codons: yield codon
